# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

from __future__ import print_function
from ..utils import clingo_stats
from ..utils import utils
import sys

BASE = utils.BASE
WARNING_INCLUDED_FILE = "<cmd>: warning: already included file:\n  {}\n"
ERROR_INCLUDED_FILE = "file could not be opened:\n  {}\n"
MESSAGE_LIMIT = 20
TOO_MANY = "too many messages."

class Printer:

    messages = 0  # class variables
    last     = "" #

    #
    # errors and warnings
    #
   
    def __check_messages(self, n):
        Printer.messages += n
        if Printer.messages >= MESSAGE_LIMIT:
            pass #raise Exception(TOO_MANY) 

    def __last(self, string):
        if Printer.last == string:
            return True
        Printer.last = string
        return False

    def __print_error(self, string, **kwargs):
        if not self.__last(string):
            sys.stdout.flush()
            print(string, file=sys.stderr, **kwargs)
            self.__check_messages(1)

    def print_error_string(self, string):
        if not self.__last(string):
            sys.stdout.flush()
            print(string, file=sys.stderr, end = "")
            self.__check_messages(int(string.count("\n")/2))

    def print_error_location(self, location, string):
        self.__print_error("{}error: {}".format(location,string))

    def print_error(self, string):
        self.__print_error(string)

    def print_spec_error(self, string):
        self.__print_error(string)

    def print_warning(self, string, **kwargs):
        if not self.__last(string):
            sys.stdout.flush()
            print(string, file=sys.stderr, **kwargs)
            self.__check_messages(1)

    def print_spec_warning(self, string):
        self.print_warning(string)

    def warning_included_file(self, file, loc=None):
        warning = WARNING_INCLUDED_FILE 
        if loc: 
            warning = warning.replace("<cmd>: ", str(loc))
        self.print_warning(warning.format(file))

    def error_included_file(self, file, loc):
        self.print_error_location(loc, ERROR_INCLUDED_FILE.format(file))

    #
    # simply print
    #
    def do_print(self, *args, **kwargs):
        print(*args, **kwargs)

    #
    # stats
    #

    def print_stats(self, ctl, models, more_models,
                    opt_models, non_optimal, stats=False):
        print("")
        print("Models       : {}{}".format(models,"+" if more_models else ""))
        if not non_optimal:
            print("  Optimum    : {}".format("yes" if opt_models>0 else "no"))
            if opt_models > 0:
                print("  Optimal    : {}".format(opt_models))
        #print(clingo_stats.Stats().summary(ctl,False))
        if stats:
            print(clingo_stats.Stats().summary(ctl,False))
            print(clingo_stats.Stats().statistics(ctl))




HELP_1="""\
clingo version 5.2.1
usage: clingo [number] [options] [files]

Clasp.Config Options:

  --configuration=<arg>   : Set default configuration [auto]
      <arg>: {auto|frumpy|jumpy|tweety|handy|crafty|trendy|many|<file>}
        auto  : Select configuration based on problem type
        frumpy: Use conservative defaults
        jumpy : Use aggressive defaults
        tweety: Use defaults geared towards asp problems
        handy : Use defaults geared towards large problems
        crafty: Use defaults geared towards crafted problems
        trendy: Use defaults geared towards industrial problems
        many  : Use default portfolio to configure solver(s)
        <file>: Use configuration file to configure solver(s)
  --tester=<options>      : Pass (quoted) string of <options> to tester
  --stats[=<n>[,<t>]],-s  : Enable {1=basic|2=full} statistics (<t> for tester)
  --[no-]parse-ext        : Enable extensions in non-aspif input
  --[no-]parse-maxsat     : Treat dimacs input as MaxSAT problem

Clasp.Solving Options:

  --parallel-mode,-t <arg>: Run parallel search with given number of threads
      <arg>: <n {1..64}>[,<mode {compete|split}>]
        <n>   : Number of threads to use in search
        <mode>: Run competition or splitting based search [compete]

  --enum-mode,-e <arg>    : Configure enumeration algorithm [auto]
      <arg>: {bt|record|brave|cautious|auto}
        bt      : Backtrack decision literals from solutions
        record  : Add nogoods for computed solutions
        domRec  : Add nogoods over true domain atoms
        brave   : Compute brave consequences (union of models)
        cautious: Compute cautious consequences (intersection of models)
        auto    : Use bt for enumeration and record for optimization
  --project[=<arg>|no]    : Enable projective solution enumeration
      <arg>: {show|project|auto}[,<bt {0..3}>] (Implicit: auto,3)
        Project to atoms in show or project directives, or
        select depending on the existence of a project directive
      <bt> : Additional options for enumeration algorithm 'bt'
        Use activity heuristic (1) when selecting backtracking literal
        and/or progress saving (2) when retracting solution literals
  --models,-n <n>         : Compute at most <n> models (0 for all)

  --opt-mode=<arg>        : Configure optimization algorithm
      <arg>: <mode {opt|enum|optN|ignore}>[,<bound>...]
        opt   : Find optimal model
        enum  : Find models with costs <= <bound>
        optN  : Find optimum, then enumerate optimal models
        ignore: Ignore optimize statements
      <bound> : Set initial bound for objective function(s)

Gringo Options:

  --text                  : Print plain text format
  --const,-c <id>=<term>  : Replace term occurrences of <id> with <term>

Basic Options:

  --help[=<n>],-h         : Print {1=basic|2=more|3=full} help and exit
  --version,-v            : Print version information and exit
  --verbose[=<n>],-V      : Set verbosity level to <n>
  --time-limit=<n>        : Set time limit to <n> seconds (0=no limit)
  --quiet[=<levels>],-q   : Configure printing of models, costs, and calls
      <levels>: <mod>[,<cost>][,<call>]
        <mod> : print {0=all|1=last|2=no} models
        <cost>: print {0=all|1=last|2=no} optimize values [<mod>]
        <call>: print {0=all|1=last|2=no} call steps      [2]
  --pre[=<fmt>]           : Print simplified program and exit
      <fmt>: Set output format to {aspif|smodels} (implicit: aspif)
  --mode=<arg>            : Run in {clingo|clasp|gringo} mode

usage: clingo [number] [options] [files]
Default command-line:
clingo --configuration=auto --enum-mode=auto --verbose=1 

Type 'clingo --help=2' for more options and defaults
and  'clingo --help=3' for all options and configurations.

clingo is part of Potassco: https://potassco.org/clingo
Get help/report bugs via : https://potassco.org/support
"""
HELP_2="""\
clingo version 5.2.1
usage: clingo [number] [options] [files]

Clasp.Config Options:

  --configuration=<arg>   : Set default configuration [auto]
      <arg>: {auto|frumpy|jumpy|tweety|handy|crafty|trendy|many|<file>}
        auto  : Select configuration based on problem type
        frumpy: Use conservative defaults
        jumpy : Use aggressive defaults
        tweety: Use defaults geared towards asp problems
        handy : Use defaults geared towards large problems
        crafty: Use defaults geared towards crafted problems
        trendy: Use defaults geared towards industrial problems
        many  : Use default portfolio to configure solver(s)
        <file>: Use configuration file to configure solver(s)
  --tester=<options>      : Pass (quoted) string of <options> to tester
  --stats[=<n>[,<t>]],-s  : Enable {1=basic|2=full} statistics (<t> for tester)
  --[no-]parse-ext        : Enable extensions in non-aspif input
  --[no-]parse-maxsat     : Treat dimacs input as MaxSAT problem

Clasp.Context Options:

  --share=<arg>|no        : Configure physical sharing of constraints [auto]
      <arg>: {auto|problem|learnt|all}
  --sat-prepro[=<arg>|no] : Run SatELite-like preprocessing (Implicit: 2)
      <arg>: <level>[,<limit>...]
        <level> : Set preprocessing level to <level  {1..3}>
          1: Variable elimination with subsumption (VE)
          2: VE with limited blocked clause elimination (BCE)
          3: Full BCE followed by VE
        <limit> : [<key {iter|occ|time|frozen|clause}>=]<n> (0=no limit)
          iter  : Set iteration limit to <n>           [0]
          occ   : Set variable occurrence limit to <n> [0]
          time  : Set time limit to <n> seconds        [0]
          frozen: Set frozen variables limit to <n>%   [0]
          size  : Set size limit to <n>*1000 clauses   [4000]

Clasp.ASP Options:

  --trans-ext=<mode>|no   : Configure handling of extended rules
      <mode>: {all|choice|card|weight|integ|dynamic}
        all    : Transform all extended rules to basic rules
        choice : Transform choice rules, but keep cardinality and weight rules
        card   : Transform cardinality rules, but keep choice and weight rules
        weight : Transform cardinality and weight rules, but keep choice rules
        scc    : Transform "recursive" cardinality and weight rules
        integ  : Transform cardinality integrity constraints
        dynamic: Transform "simple" extended rules, but keep more complex ones
  --eq=<n>                : Configure equivalence preprocessing
      Run for at most <n> iterations (-1=run to fixpoint)
  --[no-]backprop         : Use backpropagation in ASP-preprocessing
  --supp-models           : Compute supported models
  --no-ufs-check          : Disable unfounded set check
  --no-gamma              : Do not add gamma rules for non-hcf disjunctions

Clasp.Solving Options:

  --solve-limit=<n>[,<m>] : Stop search after <n> conflicts or <m> restarts

  --parallel-mode,-t <arg>: Run parallel search with given number of threads
      <arg>: <n {1..64}>[,<mode {compete|split}>]
        <n>   : Number of threads to use in search
        <mode>: Run competition or splitting based search [compete]

  --global-restarts=<X>   : Configure global restart policy
      <X>: <n>[,<sched>]
        <n> : Maximal number of global restarts (0=disable)
     <sched>: Restart schedule [x,100,1.5] (<type {F|L|x|+}>)

  --distribute=<arg>|no   : Configure nogood distribution [conflict,global,4]
      <arg>: <type>[,<mode>][,<lbd {0..127}>][,<size>]
        <type> : Distribute {all|short|conflict|loop} nogoods
        <mode> : Use {global|local} distribution   [global]
        <lbd>  : Distribute only if LBD  <= <lbd>  [4]
        <size> : Distribute only if size <= <size> [-1]
  --integrate=<arg>       : Configure nogood integration [gp]
      <arg>: <pick>[,<n>][,<topo>]
        <pick>: Add {all|unsat|gp(unsat wrt guiding path)|active} nogoods
        <n>   : Always keep at least last <n> integrated nogoods   [1024]
        <topo>: Accept nogoods from {all|ring|cube|cubex} peers    [all]

  --enum-mode,-e <arg>    : Configure enumeration algorithm [auto]
      <arg>: {bt|record|brave|cautious|auto}
        bt      : Backtrack decision literals from solutions
        record  : Add nogoods for computed solutions
        domRec  : Add nogoods over true domain atoms
        brave   : Compute brave consequences (union of models)
        cautious: Compute cautious consequences (intersection of models)
        auto    : Use bt for enumeration and record for optimization
  --project[=<arg>|no]    : Enable projective solution enumeration
      <arg>: {show|project|auto}[,<bt {0..3}>] (Implicit: auto,3)
        Project to atoms in show or project directives, or
        select depending on the existence of a project directive
      <bt> : Additional options for enumeration algorithm 'bt'
        Use activity heuristic (1) when selecting backtracking literal
        and/or progress saving (2) when retracting solution literals
  --models,-n <n>         : Compute at most <n> models (0 for all)

  --opt-mode=<arg>        : Configure optimization algorithm
      <arg>: <mode {opt|enum|optN|ignore}>[,<bound>...]
        opt   : Find optimal model
        enum  : Find models with costs <= <bound>
        optN  : Find optimum, then enumerate optimal models
        ignore: Ignore optimize statements
      <bound> : Set initial bound for objective function(s)

Clasp.Search Options:

  --opt-strategy=<arg>    : Configure optimization strategy
      <arg>: {bb|usc}[,<tactics>]
        bb : Model-guided optimization with <tactics {lin|hier|inc|dec}> [lin]
          lin : Basic lexicographical descent
          hier: Hierarchical (highest priority criteria first) descent 
          inc : Hierarchical descent with exponentially increasing steps
          dec : Hierarchical descent with exponentially decreasing steps
        usc: Core-guided optimization with <tactics>: <relax>[,<opts>]
          <relax>: Relaxation algorithm {oll|one|k|pmres}                [oll]
            oll    : Use strategy from unclasp
            one    : Add one cardinality constraint per core
            k[,<n>]: Add cardinality constraints of bounded size ([0]=dynamic)
            pmres  : Add clauses of size 3
          <opts> : Tactics <list {disjoint|succinct|stratify}>|<mask {0..7}>
            disjoint: Disjoint-core preprocessing                    (1)
            succinct: No redundant (symmetry) constraints            (2)
            stratify: Stratification heuristic for handling weights  (4)
  --opt-usc-shrink=<arg>  : Enable core-shrinking in core-guided optimization
      <arg>: <algo>[,<limit> (0=no limit)]
        <algo> : Use algorithm {lin|inv|bin|rgs|exp|min}
          lin  : Forward linear search unsat
          inv  : Inverse linear search not unsat
          bin  : Binary search
          rgs  : Repeated geometric sequence until unsat
          exp  : Exponential search until unsat
          min  : Linear search for subset minimal core
        <limit>: Limit solve calls to 2^<n> conflicts [10]
  --opt-heuristic=<list>  : Use opt. in <list {sign|model}> heuristics
  --[no-]restart-on-model : Restart after each model

  --lookahead[=<arg>|no]  : Configure failed-literal detection (fld)
      <arg>: <type>[,<limit>] / Implicit: atom
        <type> : Run fld via {atom|body|hybrid} lookahead
        <limit>: Disable fld after <limit> applications ([0]=no limit)
      --lookahead=atom is default if --no-lookback is used

  --heuristic=<heu>       : Configure decision heuristic
      <heu>: {Berkmin|Vmtf|Vsids|Domain|Unit|None}[,<n>]
        Berkmin: Use BerkMin-like heuristic (Check last <n> nogoods [0]=all)
        Vmtf   : Use Siege-like heuristic (Move <n> literals to the front [8])
        Vsids  : Use Chaff-like heuristic (Use 1.0/0.<n> as decay factor  [95])
        Domain : Use domain knowledge in Vsids-like heuristic
        Unit   : Use Smodels-like heuristic (Default if --no-lookback)
        None   : Select the first free variable
  --sign-def=<sign>       : Default sign: {asp|pos|neg|rnd}
  --dom-mod=<arg>         : Default modification for domain heuristic
      <arg>: (no|<mod>[,<pick>])
        <mod>  : Modifier {level|pos|true|neg|false|init|factor}
        <pick> : Apply <mod> to (all | <list {scc|hcc|disj|opt|show}>) atoms
  --save-progress[=<n>]   : Use RSat-like progress saving on backjumps > <n>
  --seed=<n>              : Set random number generator's seed to <n>
  --partial-check[=<arg>] : Configure partial stability tests
      <arg>: <p>[,<h>] / Implicit: 50
        <p>: Partial check skip percentage
        <h>: Init/update value for high bound ([0]=umax)
  --rand-freq=<p>|no      : Make random decisions with probability <p>
  --rand-prob=<n>[,<m>]   : Do <n> random searches with [<m>=100] conflicts

Clasp.Lookback Options:

  --no-lookback           : Disable all lookback strategies

  --forget-on-step=<opts> : Configure forgetting on (incremental) step
      <opts>: <list {varScores|signs|lemmaScores|lemmas}>|<mask {0..15}>

  --strengthen=<X>|no     : Use MiniSAT-like conflict nogood strengthening
      <X>: <mode>[,<type>][,<bump {yes|no}>]
        <mode>: Use {local|recursive} self-subsumption check
        <type>: Follow {all|short|binary} antecedents [all]
        <bump>: Bump activities of antecedents        [yes]
  --otfs[={0..2}]         : Enable {1=partial|2=full} on-the-fly subsumption
  --reverse-arcs[={0..3}] : Enable ManySAT-like inverse-arc learning
  --loops=<type>          : Configure learning of loop nogoods
      <type>: {common|distinct|shared|no}
        common  : Create loop nogoods for atoms in an unfounded set
        distinct: Create distinct loop nogood for each atom in an unfounded set
        shared  : Create loop formula for a whole unfounded set
        no      : Do not learn loop formulas

  --restarts,-r <sched>|no: Configure restart policy
      <sched>: <type {D|F|L|x|+}>,<n {1..umax}>[,<args>][,<lim>]
        F,<n>    : Run fixed sequence of <n> conflicts
        L,<n>    : Run Luby et al.'s sequence with unit length <n>
        x,<n>,<f>: Run geometric seq. of <n>*(<f>^i) conflicts  (<f> >= 1.0)
        +,<n>,<m>: Run arithmetic seq. of <n>+(<m>*i) conflicts (<m {0..umax}>)
        ...,<lim>: Repeat seq. every <lim>+j restarts           (<type> != F)
        D,<n>,<f>: Restart based on moving LBD average over last <n> conflicts
                   Mavg(<n>,LBD)*<f> > avg(LBD)
                   use conflict level average if <lim> > 0 and avg(LBD) > <lim>
      no|0       : Disable restarts
  --[no-]local-restarts   : Use Ryvchin et al.'s local restarts
  --counter-restarts=<arg>: Use counter implication restarts
      <arg>: (<rate>[,<bump>] | {0|no})
      <rate>: Interval in number of restarts
      <bump>: Bump factor applied to indegrees
  --block-restarts=<arg>  : Use glucose-style blocking restarts
      <arg>: <n>[,<R {1.0..5.0}>][,<c>]
        <n>: Window size for moving average (0=disable blocking)
        <R>: Block restart if assignment > average * <R>  [1.4]
        <c>: Disable blocking for the first <c> conflicts [10000]

  --shuffle=<n1>,<n2>|no  : Shuffle problem after <n1>+(<n2>*i) restarts

  --deletion,-d <arg>|no  : Configure deletion algorithm [basic,75,0]
      <arg>: <algo>[,<n {1..100}>][,<sc>]
        <algo>: Use {basic|sort|ipSort|ipHeap} algorithm
        <n>   : Delete at most <n>% of nogoods on reduction    [75]
        <sc>  : Use {activity|lbd|mixed} nogood scores    [activity]
      no      : Disable nogood deletion
  --del-grow=<arg>|no     : Configure size-based deletion policy
      <arg>: <f>[,<g>][,<sched>] (<f> >= 1.0)
        <f>     : Keep at most T = X*(<f>^i) learnt nogoods with X being the
                  initial limit and i the number of times <sched> fired
        <g>     : Stop growth once T > P*<g> (0=no limit)      [3.0]
        <sched> : Set grow schedule (<type {F|L|x|+}>) [grow on restart]
  --del-cfl=<sched>|no    : Configure conflict-based deletion policy
      <sched>:   <type {F|L|x|+}>,<args>... (see restarts)
  --del-init=<arg>        : Configure initial deletion limit
      <arg>: <f>[,<n>,<o>] (<f> > 0)
        <f>    : Set initial limit to P=estimated problem size/<f> [3.0]
        <n>,<o>: Clamp initial limit to the range [<n>,<n>+<o>]
  --del-estimate[=0..3]   : Use estimated problem complexity in limits
  --del-max=<n>,<X>|no    : Keep at most <n> learnt nogoods taking up to <X> MB
  --del-glue=<arg>        : Configure glue clause handling
      <arg>: <n {0..15}>[,<m {0|1}>]
        <n>: Do not delete nogoods with LBD <= <n>
        <m>: Count (0) or ignore (1) glue clauses in size limit [0]
  --del-on-restart=<n>    : Delete <n>% of learnt nogoods on each restart

Gringo Options:

  --text                  : Print plain text format
  --const,-c <id>=<term>  : Replace term occurrences of <id> with <term>
  --output,-o <arg>       : Choose output format:
      intermediate: print intermediate format
      text        : print plain text format
      reify       : print program as reified facts
      smodels     : print smodels format
                    (only supports basic features)
  --output-debug=<arg>    : Print debug information during output:
      none     : no additional info
      text     : print rules as plain text (prefix %)
      translate: print translated rules as plain text (prefix %%)
      all      : combines text and translate
  --warn,-W <warn>        : Enable/disable warnings:
      none:                     disable all warnings
      all:                      enable all warnings
      [no-]atom-undefined:      a :- b.
      [no-]file-included:       #include "a.lp". #include "a.lp".
      [no-]operation-undefined: p(1/0).
      [no-]variable-unbounded:  $x > 10.
      [no-]global-variable:     :- #count { X } = 1, X = 1.
      [no-]other:               clasp related and uncategorized warnings
  --rewrite-minimize      : Rewrite minimize constraints into rules
  --keep-facts            : Do not remove facts from normal rules
  --reify-sccs            : Calculate SCCs for reified output
  --reify-steps           : Add step numbers to reified output

Basic Options:

  --help[=<n>],-h         : Print {1=basic|2=more|3=full} help and exit
  --version,-v            : Print version information and exit
  --verbose[=<n>],-V      : Set verbosity level to <n>
  --time-limit=<n>        : Set time limit to <n> seconds (0=no limit)
  --fast-exit             : Force fast exit (do not call dtors)
  --print-portfolio       : Print default portfolio and exit
  --quiet[=<levels>],-q   : Configure printing of models, costs, and calls
      <levels>: <mod>[,<cost>][,<call>]
        <mod> : print {0=all|1=last|2=no} models
        <cost>: print {0=all|1=last|2=no} optimize values [<mod>]
        <call>: print {0=all|1=last|2=no} call steps      [2]
  --pre[=<fmt>]           : Print simplified program and exit
      <fmt>: Set output format to {aspif|smodels} (implicit: aspif)
  --outf=<n>              : Use {0=default|1=competition|2=JSON|3=no} output
  --out-hide-aux          : Hide auxiliary atoms in answers
  --lemma-in=<file>       : Read additional lemmas from <file>
  --lemma-out=<file>      : Log learnt lemmas to <file>
  --mode=<arg>            : Run in {clingo|clasp|gringo} mode

usage: clingo [number] [options] [files]
Default command-line:
clingo --configuration=auto --share=auto --distribute=conflict,global,4 
       --integrate=gp --enum-mode=auto --deletion=basic,75,0 --del-init=3.0 
       --verbose=1 
[asp] --configuration=tweety
[cnf] --configuration=trendy
[opb] --configuration=trendy

Type  'clingo --help=3' for all options and configurations.

clingo is part of Potassco: https://potassco.org/clingo
Get help/report bugs via : https://potassco.org/support
"""
HELP_3="""\
clingo version 5.2.1
usage: clingo [number] [options] [files]

Clasp.Config Options:

  --configuration=<arg>   : Set default configuration [auto]
      <arg>: {auto|frumpy|jumpy|tweety|handy|crafty|trendy|many|<file>}
        auto  : Select configuration based on problem type
        frumpy: Use conservative defaults
        jumpy : Use aggressive defaults
        tweety: Use defaults geared towards asp problems
        handy : Use defaults geared towards large problems
        crafty: Use defaults geared towards crafted problems
        trendy: Use defaults geared towards industrial problems
        many  : Use default portfolio to configure solver(s)
        <file>: Use configuration file to configure solver(s)
  --tester=<options>      : Pass (quoted) string of <options> to tester
  --stats[=<n>[,<t>]],-s  : Enable {1=basic|2=full} statistics (<t> for tester)
  --[no-]parse-ext        : Enable extensions in non-aspif input
  --[no-]parse-maxsat     : Treat dimacs input as MaxSAT problem

Clasp.Context Options:

  --share=<arg>|no        : Configure physical sharing of constraints [auto]
      <arg>: {auto|problem|learnt|all}
  --learn-explicit        : Do not use Short Implication Graph for learning
  --sat-prepro[=<arg>|no] : Run SatELite-like preprocessing (Implicit: 2)
      <arg>: <level>[,<limit>...]
        <level> : Set preprocessing level to <level  {1..3}>
          1: Variable elimination with subsumption (VE)
          2: VE with limited blocked clause elimination (BCE)
          3: Full BCE followed by VE
        <limit> : [<key {iter|occ|time|frozen|clause}>=]<n> (0=no limit)
          iter  : Set iteration limit to <n>           [0]
          occ   : Set variable occurrence limit to <n> [0]
          time  : Set time limit to <n> seconds        [0]
          frozen: Set frozen variables limit to <n>%   [0]
          size  : Set size limit to <n>*1000 clauses   [4000]

Clasp.ASP Options:

  --trans-ext=<mode>|no   : Configure handling of extended rules
      <mode>: {all|choice|card|weight|integ|dynamic}
        all    : Transform all extended rules to basic rules
        choice : Transform choice rules, but keep cardinality and weight rules
        card   : Transform cardinality rules, but keep choice and weight rules
        weight : Transform cardinality and weight rules, but keep choice rules
        scc    : Transform "recursive" cardinality and weight rules
        integ  : Transform cardinality integrity constraints
        dynamic: Transform "simple" extended rules, but keep more complex ones
  --eq=<n>                : Configure equivalence preprocessing
      Run for at most <n> iterations (-1=run to fixpoint)
  --[no-]backprop         : Use backpropagation in ASP-preprocessing
  --supp-models           : Compute supported models
  --no-ufs-check          : Disable unfounded set check
  --no-gamma              : Do not add gamma rules for non-hcf disjunctions
  --eq-dfs                : Enable df-order in eq-preprocessing

Clasp.Solving Options:

  --solve-limit=<n>[,<m>] : Stop search after <n> conflicts or <m> restarts

  --parallel-mode,-t <arg>: Run parallel search with given number of threads
      <arg>: <n {1..64}>[,<mode {compete|split}>]
        <n>   : Number of threads to use in search
        <mode>: Run competition or splitting based search [compete]

  --global-restarts=<X>   : Configure global restart policy
      <X>: <n>[,<sched>]
        <n> : Maximal number of global restarts (0=disable)
     <sched>: Restart schedule [x,100,1.5] (<type {F|L|x|+}>)

  --distribute=<arg>|no   : Configure nogood distribution [conflict,global,4]
      <arg>: <type>[,<mode>][,<lbd {0..127}>][,<size>]
        <type> : Distribute {all|short|conflict|loop} nogoods
        <mode> : Use {global|local} distribution   [global]
        <lbd>  : Distribute only if LBD  <= <lbd>  [4]
        <size> : Distribute only if size <= <size> [-1]
  --integrate=<arg>       : Configure nogood integration [gp]
      <arg>: <pick>[,<n>][,<topo>]
        <pick>: Add {all|unsat|gp(unsat wrt guiding path)|active} nogoods
        <n>   : Always keep at least last <n> integrated nogoods   [1024]
        <topo>: Accept nogoods from {all|ring|cube|cubex} peers    [all]

  --enum-mode,-e <arg>    : Configure enumeration algorithm [auto]
      <arg>: {bt|record|brave|cautious|auto}
        bt      : Backtrack decision literals from solutions
        record  : Add nogoods for computed solutions
        domRec  : Add nogoods over true domain atoms
        brave   : Compute brave consequences (union of models)
        cautious: Compute cautious consequences (intersection of models)
        auto    : Use bt for enumeration and record for optimization
  --project[=<arg>|no]    : Enable projective solution enumeration
      <arg>: {show|project|auto}[,<bt {0..3}>] (Implicit: auto,3)
        Project to atoms in show or project directives, or
        select depending on the existence of a project directive
      <bt> : Additional options for enumeration algorithm 'bt'
        Use activity heuristic (1) when selecting backtracking literal
        and/or progress saving (2) when retracting solution literals
  --models,-n <n>         : Compute at most <n> models (0 for all)

  --opt-mode=<arg>        : Configure optimization algorithm
      <arg>: <mode {opt|enum|optN|ignore}>[,<bound>...]
        opt   : Find optimal model
        enum  : Find models with costs <= <bound>
        optN  : Find optimum, then enumerate optimal models
        ignore: Ignore optimize statements
      <bound> : Set initial bound for objective function(s)

Clasp.Search Options:

  --opt-strategy=<arg>    : Configure optimization strategy
      <arg>: {bb|usc}[,<tactics>]
        bb : Model-guided optimization with <tactics {lin|hier|inc|dec}> [lin]
          lin : Basic lexicographical descent
          hier: Hierarchical (highest priority criteria first) descent 
          inc : Hierarchical descent with exponentially increasing steps
          dec : Hierarchical descent with exponentially decreasing steps
        usc: Core-guided optimization with <tactics>: <relax>[,<opts>]
          <relax>: Relaxation algorithm {oll|one|k|pmres}                [oll]
            oll    : Use strategy from unclasp
            one    : Add one cardinality constraint per core
            k[,<n>]: Add cardinality constraints of bounded size ([0]=dynamic)
            pmres  : Add clauses of size 3
          <opts> : Tactics <list {disjoint|succinct|stratify}>|<mask {0..7}>
            disjoint: Disjoint-core preprocessing                    (1)
            succinct: No redundant (symmetry) constraints            (2)
            stratify: Stratification heuristic for handling weights  (4)
  --opt-usc-shrink=<arg>  : Enable core-shrinking in core-guided optimization
      <arg>: <algo>[,<limit> (0=no limit)]
        <algo> : Use algorithm {lin|inv|bin|rgs|exp|min}
          lin  : Forward linear search unsat
          inv  : Inverse linear search not unsat
          bin  : Binary search
          rgs  : Repeated geometric sequence until unsat
          exp  : Exponential search until unsat
          min  : Linear search for subset minimal core
        <limit>: Limit solve calls to 2^<n> conflicts [10]
  --opt-heuristic=<list>  : Use opt. in <list {sign|model}> heuristics
  --[no-]restart-on-model : Restart after each model

  --lookahead[=<arg>|no]  : Configure failed-literal detection (fld)
      <arg>: <type>[,<limit>] / Implicit: atom
        <type> : Run fld via {atom|body|hybrid} lookahead
        <limit>: Disable fld after <limit> applications ([0]=no limit)
      --lookahead=atom is default if --no-lookback is used

  --heuristic=<heu>       : Configure decision heuristic
      <heu>: {Berkmin|Vmtf|Vsids|Domain|Unit|None}[,<n>]
        Berkmin: Use BerkMin-like heuristic (Check last <n> nogoods [0]=all)
        Vmtf   : Use Siege-like heuristic (Move <n> literals to the front [8])
        Vsids  : Use Chaff-like heuristic (Use 1.0/0.<n> as decay factor  [95])
        Domain : Use domain knowledge in Vsids-like heuristic
        Unit   : Use Smodels-like heuristic (Default if --no-lookback)
        None   : Select the first free variable
  --[no-]init-moms        : Initialize heuristic with MOMS-score
  --score-res=<score>     : Resolution score {auto|min|set|multiset}
  --score-other=<arg>     : Score other learnt nogoods: {auto|no|loop|all}
  --sign-def=<sign>       : Default sign: {asp|pos|neg|rnd}
  --[no-]sign-fix         : Disable sign heuristics and use default signs only
  --[no-]berk-huang       : Enable Huang-scoring in Berkmin
  --[no-]vsids-acids      : Enable acids-scheme in Vsids/Domain
  --vsids-progress=<arg>  : Enable dynamic decaying scheme in Vsids/Domain
      <arg>: <n>[,<i {1..100}>][,<c>]|(0=disable)
        <n> : Set initial decay factor to 1.0/0.<n>
        <i> : Set decay update to <i>/100.0      [1]
        <c> : Decrease decay every <c> conflicts [5000]
  --[no-]nant             : Prefer negative antecedents of P in heuristic
  --dom-mod=<arg>         : Default modification for domain heuristic
      <arg>: (no|<mod>[,<pick>])
        <mod>  : Modifier {level|pos|true|neg|false|init|factor}
        <pick> : Apply <mod> to (all | <list {scc|hcc|disj|opt|show}>) atoms
  --save-progress[=<n>]   : Use RSat-like progress saving on backjumps > <n>
  --init-watches=<arg>    : Watched literal initialization: {rnd|first|least}
  --update-mode=<mode>    : Process messages on {propagate|conflict}
  --acyc-prop[={0..1}]    : Use backward inference in acyc propagation
  --seed=<n>              : Set random number generator's seed to <n>
  --partial-check[=<arg>] : Configure partial stability tests
      <arg>: <p>[,<h>] / Implicit: 50
        <p>: Partial check skip percentage
        <h>: Init/update value for high bound ([0]=umax)
  --sign-def-disj=<sign>  : Default sign for atoms in disjunctions
  --rand-freq=<p>|no      : Make random decisions with probability <p>
  --rand-prob=<n>[,<m>]   : Do <n> random searches with [<m>=100] conflicts

Clasp.Lookback Options:

  --no-lookback           : Disable all lookback strategies

  --forget-on-step=<opts> : Configure forgetting on (incremental) step
      <opts>: <list {varScores|signs|lemmaScores|lemmas}>|<mask {0..15}>

  --strengthen=<X>|no     : Use MiniSAT-like conflict nogood strengthening
      <X>: <mode>[,<type>][,<bump {yes|no}>]
        <mode>: Use {local|recursive} self-subsumption check
        <type>: Follow {all|short|binary} antecedents [all]
        <bump>: Bump activities of antecedents        [yes]
  --otfs[={0..2}]         : Enable {1=partial|2=full} on-the-fly subsumption
  --update-lbd=<arg>|no   : Configure LBD updates during conflict resolution
      <arg>: <mode {less|glucose|pseudo}>[,<n {0..127}>]
        less   : update to X = new LBD   iff X   < previous LBD
        glucose: update to X = new LBD   iff X+1 < previous LBD
        pseudo : update to X = new LBD+1 iff X   < previous LBD
           <n> : Protect updated nogoods on next reduce if X <= <n>
  --update-act            : Enable LBD-based activity bumping
  --reverse-arcs[={0..3}] : Enable ManySAT-like inverse-arc learning
  --contraction=<arg>|no  : Configure handling of long learnt nogoods
      <arg>: <n>[,<rep>]
        <n>  : Contract nogoods if size > <n> (0=disable)
        <rep>: Nogood replacement {no|decisionSeq|allUIP|dynamic} [no]

  --loops=<type>          : Configure learning of loop nogoods
      <type>: {common|distinct|shared|no}
        common  : Create loop nogoods for atoms in an unfounded set
        distinct: Create distinct loop nogood for each atom in an unfounded set
        shared  : Create loop formula for a whole unfounded set
        no      : Do not learn loop formulas

  --restarts,-r <sched>|no: Configure restart policy
      <sched>: <type {D|F|L|x|+}>,<n {1..umax}>[,<args>][,<lim>]
        F,<n>    : Run fixed sequence of <n> conflicts
        L,<n>    : Run Luby et al.'s sequence with unit length <n>
        x,<n>,<f>: Run geometric seq. of <n>*(<f>^i) conflicts  (<f> >= 1.0)
        +,<n>,<m>: Run arithmetic seq. of <n>+(<m>*i) conflicts (<m {0..umax}>)
        ...,<lim>: Repeat seq. every <lim>+j restarts           (<type> != F)
        D,<n>,<f>: Restart based on moving LBD average over last <n> conflicts
                   Mavg(<n>,LBD)*<f> > avg(LBD)
                   use conflict level average if <lim> > 0 and avg(LBD) > <lim>
      no|0       : Disable restarts
  --reset-restarts=<arg>  : Update restart seq. on model {no|repeat|disable}
  --[no-]local-restarts   : Use Ryvchin et al.'s local restarts
  --counter-restarts=<arg>: Use counter implication restarts
      <arg>: (<rate>[,<bump>] | {0|no})
      <rate>: Interval in number of restarts
      <bump>: Bump factor applied to indegrees
  --block-restarts=<arg>  : Use glucose-style blocking restarts
      <arg>: <n>[,<R {1.0..5.0}>][,<c>]
        <n>: Window size for moving average (0=disable blocking)
        <R>: Block restart if assignment > average * <R>  [1.4]
        <c>: Disable blocking for the first <c> conflicts [10000]

  --shuffle=<n1>,<n2>|no  : Shuffle problem after <n1>+(<n2>*i) restarts

  --deletion,-d <arg>|no  : Configure deletion algorithm [basic,75,0]
      <arg>: <algo>[,<n {1..100}>][,<sc>]
        <algo>: Use {basic|sort|ipSort|ipHeap} algorithm
        <n>   : Delete at most <n>% of nogoods on reduction    [75]
        <sc>  : Use {activity|lbd|mixed} nogood scores    [activity]
      no      : Disable nogood deletion
  --del-grow=<arg>|no     : Configure size-based deletion policy
      <arg>: <f>[,<g>][,<sched>] (<f> >= 1.0)
        <f>     : Keep at most T = X*(<f>^i) learnt nogoods with X being the
                  initial limit and i the number of times <sched> fired
        <g>     : Stop growth once T > P*<g> (0=no limit)      [3.0]
        <sched> : Set grow schedule (<type {F|L|x|+}>) [grow on restart]
  --del-cfl=<sched>|no    : Configure conflict-based deletion policy
      <sched>:   <type {F|L|x|+}>,<args>... (see restarts)
  --del-init=<arg>        : Configure initial deletion limit
      <arg>: <f>[,<n>,<o>] (<f> > 0)
        <f>    : Set initial limit to P=estimated problem size/<f> [3.0]
        <n>,<o>: Clamp initial limit to the range [<n>,<n>+<o>]
  --del-estimate[=0..3]   : Use estimated problem complexity in limits
  --del-max=<n>,<X>|no    : Keep at most <n> learnt nogoods taking up to <X> MB
  --del-glue=<arg>        : Configure glue clause handling
      <arg>: <n {0..15}>[,<m {0|1}>]
        <n>: Do not delete nogoods with LBD <= <n>
        <m>: Count (0) or ignore (1) glue clauses in size limit [0]
  --del-on-restart=<n>    : Delete <n>% of learnt nogoods on each restart

Gringo Options:

  --text                  : Print plain text format
  --const,-c <id>=<term>  : Replace term occurrences of <id> with <term>
  --output,-o <arg>       : Choose output format:
      intermediate: print intermediate format
      text        : print plain text format
      reify       : print program as reified facts
      smodels     : print smodels format
                    (only supports basic features)
  --output-debug=<arg>    : Print debug information during output:
      none     : no additional info
      text     : print rules as plain text (prefix %)
      translate: print translated rules as plain text (prefix %%)
      all      : combines text and translate
  --warn,-W <warn>        : Enable/disable warnings:
      none:                     disable all warnings
      all:                      enable all warnings
      [no-]atom-undefined:      a :- b.
      [no-]file-included:       #include "a.lp". #include "a.lp".
      [no-]operation-undefined: p(1/0).
      [no-]variable-unbounded:  $x > 10.
      [no-]global-variable:     :- #count { X } = 1, X = 1.
      [no-]other:               clasp related and uncategorized warnings
  --rewrite-minimize      : Rewrite minimize constraints into rules
  --keep-facts            : Do not remove facts from normal rules
  --reify-sccs            : Calculate SCCs for reified output
  --reify-steps           : Add step numbers to reified output

Basic Options:

  --help[=<n>],-h         : Print {1=basic|2=more|3=full} help and exit
  --version,-v            : Print version information and exit
  --verbose[=<n>],-V      : Set verbosity level to <n>
  --time-limit=<n>        : Set time limit to <n> seconds (0=no limit)
  --fast-exit             : Force fast exit (do not call dtors)
  --print-portfolio       : Print default portfolio and exit
  --quiet[=<levels>],-q   : Configure printing of models, costs, and calls
      <levels>: <mod>[,<cost>][,<call>]
        <mod> : print {0=all|1=last|2=no} models
        <cost>: print {0=all|1=last|2=no} optimize values [<mod>]
        <call>: print {0=all|1=last|2=no} call steps      [2]
  --pre[=<fmt>]           : Print simplified program and exit
      <fmt>: Set output format to {aspif|smodels} (implicit: aspif)
  --outf=<n>              : Use {0=default|1=competition|2=JSON|3=no} output
  --out-atomf=<arg>       : Set atom format string (<Pre>?%0<Post>?)
  --out-ifs=<arg>         : Set internal field separator
  --out-hide-aux          : Hide auxiliary atoms in answers
  --lemma-in=<file>       : Read additional lemmas from <file>
  --lemma-out=<file>      : Log learnt lemmas to <file>
  --lemma-out-lbd=<n>     : Only log lemmas with lbd <= <n>
  --lemma-out-max=<n>     : Stop logging after <n> lemmas
  --lemma-out-dom=<arg>   : Log lemmas over <arg {input|output}> variables
  --lemma-out-txt         : Log lemmas as ground integrity constraints
  --hcc-out=<file>        : Write non-hcf programs to <file>.#scc
  --compute=<lit>         : Force given literal to true
  --mode=<arg>            : Run in {clingo|clasp|gringo} mode

usage: clingo [number] [options] [files]
Default command-line:
clingo --configuration=auto --share=auto --distribute=conflict,global,4 
       --integrate=gp --enum-mode=auto --deletion=basic,75,0 --del-init=3.0 
       --verbose=1 
[asp] --configuration=tweety
[cnf] --configuration=trendy
[opb] --configuration=trendy

Default configurations:
[tweety]:
 --eq=3 --trans-ext=dynamic --heuristic=Vsids,92 --restarts=L,60
 --deletion=basic,50 --del-max=2000000 --del-estimate=1 --del-cfl=+,2000,100,20
 --del-grow=0 --del-glue=2,0 --strengthen=recursive,all --otfs=2 --init-moms
 --score-other=all --update-lbd=less --save-progress=160 --init-watches=least
 --local-restarts --loops=shared
[trendy]:
 --sat-p=2,iter=20,occ=25,time=240 --trans-ext=dynamic --heuristic=Vsids
 --restarts=D,100,0.7 --deletion=basic,50 --del-init=3.0,500,19500
 --del-grow=1.1,20.0,x,100,1.5 --del-cfl=+,10000,2000 --del-glue=2
 --strengthen=recursive --update-lbd=less --otfs=2 --save-p=75
 --counter-restarts=3,1023 --reverse-arcs=2 --contraction=250 --loops=common
[frumpy]:
 --eq=5 --heuristic=Berkmin --restarts=x,100,1.5 --deletion=basic,75
 --del-init=3.0,200,40000 --del-max=400000 --contraction=250 --loops=common
 --save-p=180 --del-grow=1.1 --strengthen=local --sign-def-disj=pos
[crafty]:
 --sat-p=2,iter=10,occ=25,time=240 --trans-ext=dynamic --backprop
 --heuristic=Vsids --save-p=180 --restarts=x,128,1.5 --deletion=basic,75
 --del-init=10.0,1000,9000 --del-grow=1.1,20.0 --del-cfl=+,10000,1000
 --del-glue=2 --otfs=2 --reverse-arcs=1 --counter-restarts=3,9973
 --contraction=250
[jumpy]:
 --sat-p=2,iter=20,occ=25,time=240 --trans-ext=dynamic --heuristic=Vsids
 --restarts=L,100 --deletion=basic,75,mixed --del-init=3.0,1000,20000
 --del-grow=1.1,25,x,100,1.5 --del-cfl=x,10000,1.1 --del-glue=2
 --update-lbd=glucose --strengthen=recursive --otfs=2 --save-p=70
[handy]:
 --sat-p=2,iter=10,occ=25,time=240 --trans-ext=dynamic --backprop
 --heuristic=Vsids --restarts=D,100,0.7 --deletion=sort,50,mixed
 --del-max=200000 --del-init=20.0,1000,14000 --del-cfl=+,4000,600 --del-glue=2
 --update-lbd=less --strengthen=recursive --otfs=2 --save-p=20
 --contraction=600 --loops=distinct --counter-restarts=7,1023 --reverse-arcs=2

clingo is part of Potassco: https://potassco.org/clingo
Get help/report bugs via : https://potassco.org/support
"""



# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

#!/usr/bin/python

import os
import sys
import clingo
import json

#
# STATS
#

class Stats:


    def __init__(self):
        self.__width = 13



    def __ratio(self,x,y):
        return float(x)/float(y) if float(y)!=0 else 0


    def __percent(self,x,y):
        return 100*self.__ratio(x,y)


    def __print_key(self,key):
        return key + " "*(self.__width-len(key)) + ": "


    def __print_key_value(self,key,value):
        return self.__print_key(key) + value


    # requires Control initialized with --stats
    def summary(self,control,models=True):

        out = ""
        summary = control.statistics['summary']
        moreStr = "+" if int(summary['exhausted'])==0 else ""
        numEnum = int(summary['models']['enumerated'])
        if models:
            out += self.__print_key("Models")
            out += "{}{}\n".format(numEnum,moreStr)

        step = int(summary['call'])
        out += self.__print_key_value("Calls","{}\n".format(step+1))

        # return out if no stats
        if not 'accu' in control.statistics: return out
        times = control.statistics['accu']['times']
        out += self.__print_key("Time")
        totalTime = float(times['total'])
        solveTime = float(times['solve'])
        satTime   = float(times['sat'])
        unsatTime = float(times['unsat'])
        cpuTime   = float(times['cpu'])
        out += "{:.3f}s (Solving: {:.2f}s 1st Model: {:.2f}s Unsat: {:.2f}s)\n".format(totalTime,solveTime,satTime,unsatTime)
        out += self.__print_key_value("CPU Time","{:.3f}s".format(cpuTime))

        concurrency = int(summary['concurrency'])
        if concurrency > 1:
            out += "\n" + self.__print_key_value("Threads","{:<8}".format(concurrency))
            # when winner info becomes available: " (Winner: {})\n".format(winner)

        return out


    # requires Control initialized with --stats
    def statistics(self,control):

        # return "" if no stats
        if not 'accu' in control.statistics: return ""

        # choices...
        solver      = control.statistics['accu']['solving']['solvers']
        extra       = solver['extra']
        choices     = int(solver['choices'])
        domChoices  = int(extra['domain_choices'])
        conflicts   = int(solver['conflicts'])
        backjumps   = int(solver['conflicts_analyzed'])
        restarts    = int(solver['restarts'])
        avgRestart  = self.__ratio(backjumps,restarts)
        lastRestart = int(solver['restarts_last'])
        out = "\n" + self.__print_key_value("Choices","{:<8}".format(choices))
        if domChoices: out += " (Domain: {})".format(domChoices)
        out += "\n"
        out += self.__print_key_value("Conflicts","{:<8}".format(conflicts))
        out += " (Analyzed: {})\n".format(backjumps)
        out += self.__print_key_value("Restarts","{:<8}".format(restarts))
        if restarts>0: out += " (Average: {:.2f} Last: {})".format(avgRestart,lastRestart)
        out += "\n"

        # hccs
        hccTests   = int(extra['hcc_tests'])
        hccPartial = int(extra['hcc_partial'])
        if hccTests:
            out += self.__print_key_value("Stab. Tests","{:<8}".format(hccTests))
            out += " (Full: {} Partial: {})\n".format(hccTests-hccPartial,hccPartial)

        # model level
        models    = extra['models']
        modelLits = extra['models_level']
        avgModel  = self.__ratio(modelLits,models)
        if models:
            out += self.__print_key_value("Model-Level","{:<8.1f}\n".format(avgModel))

        # lemmas
        gps           = int(extra['guiding_paths'])
        gpLits        = int(extra['guiding_paths_lits'])
        avgGp         = self.__ratio(gpLits, gps)
        splits        = int(extra['splits'])
        sum           = int(extra['lemmas'])
        deleted       = int(extra['lemmas_deleted'])
        binary        = int(extra['lemmas_binary'])
        ternary       = int(extra['lemmas_ternary'])
        conflict      = int(extra['lemmas_conflict'])
        loop          = int(extra['lemmas_loop'])
        other         = int(extra['lemmas_other'])
        lits_conflict = int(extra['lits_conflict'])
        lits_loop     = int(extra['lits_loop'])
        lits_other    = int(extra['lits_other'])
        out += self.__print_key_value("Problems","{:<8}".format(gps))
        out += " (Average Length: {:.2f} Splits: {})\n".format(avgGp,splits)
        out += self.__print_key_value("Lemmas","{:<8}".format(sum))
        out += " (Deleted: {})\n".format(deleted)
        out += self.__print_key_value("  Binary","{:<8}".format(binary))
        out += " (Ratio: {:6.2f}%)\n".format(self.__percent(binary,sum))
        out += self.__print_key_value("  Ternary","{:<8}".format(ternary))
        out += " (Ratio: {:6.2f}%)\n".format(self.__percent(ternary,sum))
        out += self.__print_key_value("  Conflict","{:<8}".format(conflict))
        out += " (Average Length: {:6.1f} Ratio: {:6.2f}%) \n".format(self.__ratio(lits_conflict,conflict),self.__percent(conflict,sum))
        out += self.__print_key_value("  Loop","{:<8}".format(loop))
        out += " (Average Length: {:6.1f} Ratio: {:6.2f}%) \n".format(self.__ratio(lits_loop,loop),self.__percent(loop,sum))
        out += self.__print_key_value("  Other","{:<8}".format(other))
        out += " (Average Length: {:6.1f} Ratio: {:6.2f}%) \n".format(self.__ratio(lits_other,other),self.__percent(other,sum))

        # distributed...
        distributed = int(extra['distributed'])
        integrated  = int(extra['integrated'])
        if distributed or integrated:
            distRatio   = self.__ratio(distributed,conflict+loop)
            sumDistLbd  = int(extra['distributed_sum_lbd'])
            avgDistLbd  = self.__ratio(sumDistLbd,distributed)
            intRatio    =  self.__ratio(integrated,distributed)
            intImps     = int(extra['integrated_imps'])
            intJumps    = int(extra['integrated_jumps'])
            avgIntJump  = self.__ratio(intJumps,intImps)
            out += self.__print_key_value("  Distributed","{:<8}".format(distributed))
            out += " (Ratio: {:6.2f}% Average LBD: {:.2f}) \n".format(distRatio*100.0,avgDistLbd)
            out += self.__print_key_value("  Integrated","{:<8}".format(integrated))
            out += " (Ratio: {:6.2f}% ".format(intRatio*100.0) # for not accu: if not _accu: "("
            out += "Unit: {} Average Jumps: {:.2f})\n".format(intImps,avgIntJump)

        # jumps
        jumps       = extra['jumps']
        _jumps      = int(jumps['jumps'])
        bounded     = int(jumps['jumps_bounded'])
        jumpSum     = int(jumps['levels'])
        boundSum    = int(jumps['levels_bounded'])
        maxJump     = int(jumps['max'])
        maxJumpEx   = int(jumps['max_executed'])
        maxBound    = int(jumps['max_bounded'])
        jumped      = jumpSum - boundSum
        jumpedRatio = self.__ratio(jumped,jumpSum)
        avgBound    = self.__ratio(boundSum,bounded)
        avgJump     = self.__ratio(jumpSum,_jumps)
        avgJumpEx   = self.__ratio(jumped,_jumps)
        out += self.__print_key_value("Backjumps","{:<8}".format(_jumps))
        out += " (Average: {:5.2f} Max: {:>3} Sum: {:>6})\n".format(avgJump,maxJump,jumpSum)
        out += self.__print_key_value("  Executed","{:<8}".format(_jumps-bounded))
        out += " (Average: {:5.2f} Max: {:>3} Sum: {:>6} Ratio: {:6.2f}%)\n".format(avgJumpEx,maxJumpEx,jumped,jumpedRatio*100.0)
        out += self.__print_key_value("  Bounded","{:<8}".format(bounded))
        out += " (Average: {:5.2f} Max: {:>3} Sum: {:>6} Ratio: {:6.2f}%)\n".format(avgBound,maxBound,boundSum,100.0 - (jumpedRatio*100.0))
        out += "\n"

        # logic program
        lp                 = control.statistics['problem']['lp']

        # rules
        rOriginal          = int(lp['rules'])
        rules_normal       = int(lp['rules_normal'])
        rules_choice       = int(lp['rules_choice'])
        rules_minimize     = int(lp['rules_minimize'])
        rules_acyc         = int(lp['rules_acyc'])
        rules_heuristic    = int(lp['rules_heuristic'])
        rFinal             = int(lp['rules_tr'])
        rules_tr_normal    = int(lp['rules_tr_normal'])
        rules_tr_choice    = int(lp['rules_tr_choice'])
        rules_tr_minimize  = int(lp['rules_tr_minimize'])
        rules_tr_acyc      = int(lp['rules_tr_acyc'])
        rules_tr_heuristic = int(lp['rules_tr_heuristic'])
        out += self.__print_key_value("Rules","{:<8}".format(rFinal))
        if (rFinal != rOriginal):
            out += " (Original: {})".format(rOriginal)
        out += "\n"
        for i in [#["  Normal",   rules_normal,   rules_tr_normal],
                  ["  Choice",   rules_choice,   rules_tr_choice],
                  ["  Minimize", rules_minimize, rules_tr_minimize],
                  ["  Acyc", rules_acyc, rules_tr_acyc],
                  ["  Heuristic", rules_heuristic, rules_tr_heuristic]]:
            if i[2]:
                out += self.__print_key_value(i[0],"{:<8}".format(i[2]))
                if (i[2] != i[1]):
                    out += " (Original: {})".format(i[1])
                out += "\n"

        # atoms
        atoms              = int(lp['atoms'])
        auxAtoms           = int(lp['atoms_aux'])
        out += self.__print_key_value("Atoms","{:<8}".format(atoms))
        if (auxAtoms):
            out += " (Original: {} Auxiliary: {})".format(atoms-auxAtoms,auxAtoms)
        out += "\n"

        # disjunctions
        disjunctions       = int(lp['disjunctions'])
        disjunctions_tr    = int(lp['disjunctions_non_hcf'])
        if disjunctions:
            out += self.__print_key_value("Disjunctions","{:<8}".format(disjunctions_tr))
            out += " (Original: {})\n".format(disjunctions)

        # bodies
        bFinal             = int(lp['bodies_tr'])
        bOriginal          = int(lp['bodies'])
        count_bodies       = int(lp['count_bodies'])
        count_bodies_tr    = int(lp['count_bodies_tr'])
        sum_bodies         = int(lp['sum_bodies'])
        sum_bodies_tr      = int(lp['sum_bodies_tr'])
        out += self.__print_key_value("Bodies","{:<8}".format(bFinal))
        if (bFinal != bOriginal):
            out += " (Original: {})".format(bOriginal)
        out += "\n"
        for i in [["  Count", count_bodies, count_bodies_tr],
                  ["  Sum",   sum_bodies,   sum_bodies_tr  ]]:
            if i[1]:
                out += self.__print_key_value(i[0],"{:<8}".format(i[2]))
                if (i[2] != i[1]):
                    out += " (Original: {})".format(i[1])
                out += "\n"

        # equivalences
        eqs                = int(lp['eqs'])
        eqsAtom            = int(lp['eqs_atom'])
        eqsBody            = int(lp['eqs_body'])
        eqsOther           = int(lp['eqs_other'])
        if eqs > 0:
            out += self.__print_key_value("Equivalences","{:<8}".format(eqs))
            out += " (Atom=Atom: {} Body=Body: {} Other: {})\n".format(eqsAtom,eqsBody,eqsOther)

        # sccs
        sccs       = int(lp['sccs'])
        nonHcfs    = int(lp['sccs_non_hcf'])
        ufsNodes  = int(lp['ufs_nodes'])
        gammas     = int(lp['gammas'])
        out += self.__print_key("Tight")
        if sccs==0: out += "Yes"
        # for supported models: elif sccs == PrgNode:noScc
        else: out += "{:<8} (SCCs: {} Non-Hcfs: {} Nodes: {} Gammas: {})".format("No",sccs,nonHcfs,ufsNodes,gammas)
        out += "\n"

        # problem
        gen        = control.statistics['problem']['generator']
        vars       = int(gen['vars'])
        eliminated = int(gen['vars_eliminated'])
        frozen     = int(gen['vars_frozen'])
        binary     = int(gen['constraints_binary'])
        ternary    = int(gen['constraints_ternary'])
        sum        = int(gen['constraints']) + binary + ternary
        acycEdges  = int(gen['acyc_edges'])
        out += self.__print_key_value("Variables","{:<8}".format(vars))
        out += " (Eliminated: {:>4} Frozen: {:>4})\n".format(eliminated,frozen)
        out += self.__print_key_value("Constraints","{:<8}".format(sum))
        out += " (Binary: {:5.1f}% Ternary: {:5.1f}% Other: {:5.1f}%)\n".format(self.__percent(binary,sum),self.__percent(ternary,sum),self.__percent(sum-binary-ternary,sum))

        return out


program = """

% pigeonhole problem
#const n=8.
pigeon(1..n+1). box(1..n).
1 { in(X,Y) : box(Y) } 1 :- pigeon(X).
:- 2 { in(X,Y) : pigeon(X) },  box(Y).

% heuristic
#heuristic in(X,Y) : pigeon(X), box(Y). [1,true]

% disjunction
a | b.  a :- b.  b :- a.

% SAT
box(n+1).

"""

satisfiable = False
def on_model(model):
    global satisfiable
    sys.stdout.write("Answer: 1\n{}\n".format(str(model)))
    satisfiable = True


def run():

    # set options
    options = "-t4 --stats --heuristic=Domain"
    #options = ""

    # with Control()
    control = clingo.Control(options.split())
    control.add("a",[],program)
    control.ground([("a",[])])
    control.solve(on_model=on_model)
    if satisfiable: sys.stdout.write("SATISFIABLE\n")
    else:           sys.stdout.write("UNSATISFIABLE\n")
    sys.stdout.write(Stats().summary(control)+"\n")
    sys.stdout.write(Stats().statistics(control)+"\n")

    # with $clingo
    file = "tmp.lp"
    with open(file, "w") as text_file:
        text_file.write(program)
    os.system("clingo {} {}; rm {}".format(options,file,file))

if __name__ == "__main__":
    run()




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

import solver as solver_module
from ..utils import utils

class GeneralController:

    def __init__(self, solver):
        solver.step         = 1
        solver.last_unsat   = True
        solver.opt_models   = 0
        solver.models       = 0
        solver.more_models  = True
        solver.old_holds    = None
        solver.old_nholds   = None
        if solver.options.max_models == 1 and not solver.options.improve_limit:
            solver.store_nholds = False
        self.improve_limit = True if solver.options.improve_limit is not None \
            else False
        self.solver         = solver

    def start(self):
        solver, options = self.solver, self.solver.options
        if options.trans_ext is not None:
            solver.control.configuration.asp.trans_ext = options.trans_ext
        solver.set_holds_domain()
        solver.add_encodings()
        solver.ground_preference_base()
        if options.cmd_heuristic is not None:
            solver.ground_cmd_heuristic()
            for _solver in solver.control.configuration.solver:
                _solver.heuristic="Domain"
        if options.preference_unsat and options.max_models != 1:
            solver.get_preference_parts_opt = solver.get_preference_parts_unsatp
            solver.ground_unsatp_base()
        # check syntax
        if options.check:
            solver.check_errors()
        # non optimal
        if solver.no_optimize():
            # modifying options
            solver.options.non_optimal = True
        if solver.options.non_optimal:
            solver.str_found      = solver_module.STR_MODEL_FOUND
            solver.str_found_star = solver_module.STR_MODEL_FOUND_STAR
            solver.add_unsat_to_preference_program()


    def sat(self):
        self.solver.models     += 1
        self.solver.last_unsat  = False
        self.solver.last_model  = self.solver.step
        self.solver.check_last_model()
        if self.solver.options.quiet == 0:
            self.solver.print_answer()
        else:
            self.solver.print_str_answer()

    def unsat(self):
        if self.solver.last_unsat:
            self.solver.more_models = False
            if self.solver.models == 0:
                self.solver.print_unsat()
            elif self.improve_limit:
                self.solver.enumerate_unknown()
            self.solver.end()
        if self.solver.options.quiet == 1:
            self.solver.print_shown()
        self.solver.last_unsat = True
        self.solver.opt_models  += 1
        self.solver.print_optimum_string()
        if self.solver.opt_models == self.solver.options.max_models:
            self.solver.end()
        if self.solver.options.steps == self.solver.step:
            self.solver.end() # to exit asap in this case

    def unknown(self):
        if self.solver.last_unsat:
            raise utils.FatalException
        self.solver.print_limit_string()
        self.solver.last_unsat = True # act as if unsat
        if self.solver.options.steps == self.solver.step:
            self.solver.end() # to exit asap in this case

    def end_loop(self):
        self.solver.step = self.solver.step + 1
        if self.solver.options.steps == (self.solver.step - 1):
            self.solver.end()
        if self.solver.options.clean_up:
            self.solver.clean_up()


class GeneralControllerHandleOptimal:

    def __init__(self, solver):
        self.solver = solver
        self.start_step = True
        self.first  = True
        self.delete_worse  = True
        self.delete_better = self.solver.options.delete_better
        self.total_order   = self.solver.options.total_order
        self.no_opt_improving, self.volatile = False, False
        if solver.options.no_opt_improving:
            if solver.options.max_models != 1:
                self.no_opt_improving = True
            self.volatile = True
        if solver.options.volatile_optimal:
            self.volatile = True

    def start(self):
        if self.delete_better:
            self.solver.ground_holds_delete_better()

    def sat(self):
        if self.no_opt_improving and self.start_step:
            self.solver.relax_optimal_models()
        self.start_step = False

    def __preprocess(self):
        if self.total_order and not self.first:
            self.delete_worse  = False
            self.delete_better = False
        self.first = False

    def unsat(self):
        self.start_step = True
        self.__preprocess()
        self.solver.handle_optimal_model(self.solver.last_model,
                                         False,
                                         self.delete_worse,
                                         self.delete_better,
                                         self.volatile)

    def unknown(self):
        self.start_step = True
        self.__preprocess()
        self.solver.handle_optimal_model(self.solver.last_model,
                                         True,
                                         self.delete_worse,
                                         False,
                                         True)


class EnumerationController:

    def __init__(self, solver):
        self.solver = solver
        # show_underscores is set by program_parser.py
        if not solver.options.project and not solver.options.show_underscores:
            solver.same_shown_function = solver.same_shown

    def unsat(self):
        if not self.solver.options.project:
            self.solver.enumerate()


#
# Method Controllers
#

class MethodController:

    def __init__(self, solver):
        self.solver = solver

    def start(self):
        pass

    def start_loop(self):
        pass

    def solve(self):
        pass

    def unsat(self):
        pass


class GroundManyMethodController(MethodController):

    def __init__(self, solver):
        MethodController.__init__(self, solver)
        self.volatile = False
        self.ground_holds = 0
        if (self.solver.options.max_models != 1 or
            self.solver.options.release_last or 
            self.solver.options.volatile_improving):
            self.volatile = True 

    def start_loop(self):
        if not self.solver.last_unsat:
            self.solver.ground_holds(self.solver.last_model)
            self.ground_holds = self.solver.last_model
            self.solver.ground_preference_program(self.volatile)

    def solve(self):
        self.solver.solve()

    def unsat(self):
        self.solver.relax_previous_models()
        if self.ground_holds != self.solver.last_model:
            self.solver.ground_holds(self.solver.last_model)


class GroundOnceMethodController(MethodController):

    def __init__(self, solver):
        MethodController.__init__(self, solver)
        solver.store_nholds = True

    def start(self):
        self.solver.ground_open_preference_program()
        self.solver.turn_off_preference_program()

    def start_loop(self):
        if not self.solver.last_unsat:
            self.solver.turn_on_preference_program()

    def solve(self):
        self.solver.solve()

    def unsat(self):
        self.solver.ground_holds(self.solver.last_model)
        self.solver.turn_off_preference_program()


class ApproxMethodController(MethodController):

    def __init__(self, solver):
        MethodController.__init__(self, solver)

    def start(self):
        self.solver.solve_approx()
        # finishes asprin


class HeurMethodController(MethodController):

    def __init__(self, solver):
        MethodController.__init__(self, solver)

    def start(self):
        self.solver.ground_heuristic()

    def solve(self):
        if self.solver.last_unsat:
            self.solver.solve_heuristic()
        else:
            self.solver.solve_unsat()

    def unsat(self):
        self.solver.ground_holds(self.solver.last_model)


class ImproveLimitController(MethodController):

    def __init__(self, solver, controller):
        MethodController.__init__(self, solver)
        self.solver     = solver
        self.controller = controller
        self.option     = self.solver.options.improve_limit
        self.stats      = solver.control.statistics['solving']['solvers']
        self.conf       = solver.control.configuration.solve
        self.search     = 0
        # limits
        self.previous_limit, self.limit = "", 0

    def start(self):
        self.controller.start()
        self.solver.ground_holds_delete_better()

    def start_loop(self):
        # get previous limit, and set limit
        self.previous_limit = self.conf.solve_limit
        if not self.solver.last_unsat:
            self.limit = self.search * self.option[0]
            if self.limit < self.option[2]:
                self.limit = self.option[2]
            self.conf.solve_limit = str(self.limit) + ",umax"
            if self.limit == 0: # if limit is 0, return
                return
        # call controller
        self.controller.start_loop()

    def solve(self):
        # if improving and limit is 0, return unknown
        if not self.solver.last_unsat and self.limit == 0:
            self.solver.solve_unknown()
        else:
            self.controller.solve()
        # reset previous limit
        self.conf.solve_limit = self.previous_limit
        # gather search results
        if self.solver.solving_result == solver_module.SATISFIABLE:
            if self.solver.last_unsat:
                self.search  = int(self.stats['conflicts'])
            elif self.option[1]:
                self.search += int(self.stats['conflicts'])

    def unsat(self):
        self.controller.unsat()
        self.solver.handle_unknown_models()






# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

import clingo.ast
from ..utils import utils
from ..program_parser import transitive_closure
from ..program_parser import visitor


#
# DEFINES
#

# predicates
HOLDS      = utils.HOLDS
HOLDSP     = utils.HOLDSP
PREFERENCE = utils.PREFERENCE
OPTIMIZE   = utils.OPTIMIZE

# graph
HOLDS_KEY  = (HOLDS,  1)
HOLDSP_KEY = (HOLDSP, 1)
OPEN_NAME  = "open"

# ems
M1 = visitor.M1
M2 = visitor.M2
M1_M2 = visitor.M1_M2

# errors
ERROR_PROJECT    = utils.ERROR_PROJECT
ERROR_MINIMIZE   = utils.ERROR_MINIMIZE
ERROR_HEURISTIC  = utils.ERROR_HEURISTIC
ERROR_DISJOINT   = utils.ERROR_DISJOINT
ERROR_CSPLITERAL = utils.ERROR_CSPLITERAL
ERROR_KEYWORD    = utils.ERROR_KEYWORD
ERROR_HOLDS_H    = utils.ERROR_HOLDS_H
ERROR_HOLDSP_H   = utils.ERROR_HOLDSP_H

# no sign
NO_SIGN = utils.NO_SIGN

# finish() function return value
SKIP   = 1 # do not add to program
DET    = 2 # add deterministic part 
NONDET = 3 # add nondeterministic part


class PreferenceTermTransformer(visitor.TermTransformer):

    def __init__(self, type, underscores):
        visitor.TermTransformer.__init__(self)
        self.__type = type
        self.__underscores = underscores

    def set_predicates_info(self, open):
        self.predicates_info = dict()
        PredicateInfo = visitor.PredicateInfo
        info = PredicateInfo(None, self.__underscores, M1_M2, 2)
        for i in open:
            self.predicates_info[i] = info
        # HOLDS and HOLDS' appear always in open, and should be always overriden
        self.predicates_info[(HOLDS, 1)]  = PredicateInfo(None, 0, M1, 1)
        self.predicates_info[(HOLDSP, 1)] = PredicateInfo(HOLDS, 0, M2, 1)
        info = PredicateInfo(None, 0, None, 0)
        for i in [(OPTIMIZE, 1), (PREFERENCE, 2), (PREFERENCE, 5)]:
            if i in self.predicates_info:
                string = ERROR_KEYWORD.format(self.__type, i[0], i[1])
                visitor.Helper().raise_exception(string)
            else:
                self.predicates_info[i] = info
        self.default = PredicateInfo(None, self.__underscores, None, 0)


class Graph:

    def __init__(self):
        self.__tc = transitive_closure.TransitiveClosure()
        NodeInfo = transitive_closure.NodeInfo
        self.__open = NodeInfo((visitor.Helper().underscore(OPEN_NAME),0), None)
        holds  = NodeInfo(HOLDS_KEY,  None)
        holdsp = NodeInfo(HOLDSP_KEY, None)
        self.__tc.add_node(self.__open)
        self.__tc.add_node(holds)
        self.__tc.add_node(holdsp)
        self.__tc.add_edge(self.__open, holds, True)
        self.__tc.add_edge(self.__open, holdsp, True)
        self.__heads, self.__bodies = [], []

    def __str__(self):
        out = str(self.__tc)
        out += "\n" + str(self.__tc.get_next(self.__open.key))
        return out

    #
    # parsing a rule:
    #   add_atom()* update()
    #

    def __get_info(self, term):
        return transitive_closure.NodeInfo((term.name, len(term.arguments)),
                                           term)

    def add_atom(self, term, in_head, in_body, flag):
        info = self.__get_info(term)
        self.__tc.add_node(info)
        if in_head:
            self.__heads.append((info, flag))
            if flag:
                self.__tc.add_edge(self.__open, info, False)
        if in_body:
            self.__bodies.append((info, flag))

    def update(self):
        for i in self.__heads:
            for j in self.__bodies:
                self.__tc.add_edge(j[0], i[0], j[1])
        self.__heads, self.__bodies = [], []

    #
    # after parsing all rules
    #

    def get_open(self):
        for i in self.__tc.get_cycles():
            self.__tc.add_edge(self.__open, i, False)
        return self.__tc.get_next(self.__open.key)

    def map_items(self, function):
        self.__tc.map_items(function)


class Statement:
    def __init__(self, type, statement):
        self.type = type
        self.statement = statement
        self.preds = []


class Condition:
    def __init__(self, condition, location):
        self.condition = condition
        self.location = location
        self.preds = []


class PreferenceProgramVisitor(visitor.Visitor):

    def __init__(self, builder, type, underscores, constants):
        visitor.Visitor.__init__(self)
        self.__builder    = builder
        self.__type       = type
        self.__constants  = constants
        self.__statements = []
        self.__conditions = []
        self.__graph      = Graph()
        self.__helper     = visitor.Helper()
        self.__term_transformer = PreferenceTermTransformer(self.__type,
                                                            underscores)
        # tracing position
        self.in_Head, self.in_Body, self.in_Condition = False, False, False
        self.in_Literal_ConditionalLiteral = False

    #
    # AUXILIARY FUNCTIONS
    #

    def __visit_body(self, body):
        self.in_Body = True
        self.visit(body)
        self.in_Body = False

    def __visit_condition(self, condition, location):
        self.in_Condition = True
        self.__conditions.append(Condition(condition, location))
        self.visit(condition)
        self.in_Condition = False

    # element is either a Statement or a Condition
    def __is_nondet(self, element, open_list):
        for i in element.preds:
            if i in open_list:
                return True
        return False

    def __add_volatile(self, body, location):
        body.append(self.__helper.get_volatile_atom(location))

    def __get_params(self, location):
        id1 = clingo.ast.Id(location, self.__helper.underscore(M1))
        id2 = clingo.ast.Id(location, self.__helper.underscore(M2))
        return [id1, id2]

    def __get_program(self, deterministic_type, loc):
        if deterministic_type == DET:
            return clingo.ast.Program(loc, utils.mapbase[self.__type], [])
        return clingo.ast.Program(loc, self.__type, self.__get_params(loc))

    #
    # finish() after visiting all
    #

    def finish(self):
        # get open predicates from the graph
        open_list = self.__graph.get_open()
        # set predicates_info in term_transformer
        self.__term_transformer.set_predicates_info(open_list)
        # tranform items stored in the graph
        self.__graph.map_items(self.__term_transformer.transform_function)
        # iterate over conditions
        for c in self.__conditions:
            if self.__is_nondet(c, open_list):
                self.__add_volatile(c.condition, c.location)
        # iterate over statements
        deterministic_type = None
        for st in self.__statements:
            method = "finish_" + st.type
            ret = getattr(self, method, lambda x, y: DET)(st, open_list)
            if ret is not SKIP:
                if ret != deterministic_type:
                    self.__builder.add(self.__get_program(
                                       ret, st.statement.location))
                    deterministic_type = ret
                self.__builder.add(st.statement)

    #
    # Statements
    #

    def visit_Rule(self, rule):
        self.__statements.append(Statement("Rule", rule))
        # head
        self.in_Head = True
        # if empty head
        if (str(rule.head.type) == "Literal"
            and str(rule.head.atom.type) == "BooleanConstant"
            and str(rule.head.atom.value == False)):
            # add unsat head
            atom = clingo.ast.SymbolicAtom(clingo.ast.Function(rule.location,
                   self.__helper.unsat,
                   self.__helper.get_ems(rule.location, M1_M2), False))
            rule.head = clingo.ast.Literal(rule.location, NO_SIGN, atom)
            self.__statements[-1].preds.append(HOLDS_KEY) # make it nondet
        else:
            self.visit(rule.head)
        self.in_Head = False
        # body
        self.__visit_body(rule.body)
        # graph
        self.__graph.update()

    def finish_Rule(self, statement, open_list):
        rule = statement.statement
        if self.__is_nondet(statement, open_list):
            self.__add_volatile(rule.body, rule.location)
            return NONDET
        return DET

    def visit_Definition(self, d):
        if d.name not in self.__constants:
            self.__statements.append(Statement("Definition", d))

    def visit_ShowSignature(self, sig):
        self.__statements.append(Statement("ShowSignature", sig))

    def finish_ShowSignature(self, statement, open_list):
        self.__term_transformer.transform_signature(statement.statement)
        return DET

    def visit_ShowTerm(self, show):
        self.__statements.append(Statement("ShowTerm", show))
        self.__visit_body(show.body)
        self.__graph.update()

    def finish_ShowTerm(self, statement, open_list):
        show, tt = statement.statement, self.__term_transformer
        show.term = tt.reify_term(show.term, self.__helper.show)
        if self.__is_nondet(statement, open_list):
            self.__add_volatile(show.body, show.location)
            tt.extend_function(show.term, M1_M2)
            return NONDET
        return DET

    def visit_Minimize(self, min):
        string = ERROR_MINIMIZE.format(self.__type, str(min))
        self.__helper.raise_exception(string)

    def visit_Script(self, script):
        self.__statements.append(Statement("Script", script))

    def visit_Program(self, prg):
        pass # ignore

    def visit_External(self, ext):
        pass # ignore

    # TODO: Warn if computing many models (as for disjunction...)
    def visit_Edge(self, edge):
        self.__statements.append(Statement("Edge", edge))
        self.__visit_body(edge.body)
        self.__graph.update()

    def finish_Edge(self, statement, open_list):
        edge, tt = statement.statement, self.__term_transformer
        edge.u = tt.reify_term(edge.u, self.__helper.edge)
        tt.extend_function(edge.u, M1_M2)
        edge.v = tt.reify_term(edge.v, self.__helper.edge)
        tt.extend_function(edge.v, M1_M2)
        # could be done only if *some* edge is nondet
        self.__add_volatile(edge.body, edge.location)
        return NONDET

    def visit_Heuristic(self, heur):
        self.__statements.append(Statement("Heuristic", heur))
        self.__visit_body(heur.body)
        # head
        term = heur.atom.term
        if str(term.type) == "Function":
            self.__graph.add_atom(term, False, False, False)
            self.__statements[-1].heur_atom = (term.name, len(term.arguments))
        # update graph
        self.__graph.update()

    def finish_Heuristic(self, statement, open_list):
        heur = statement.statement
        # not open -> skip
        if statement.heur_atom not in open_list:
            return SKIP
        # holds
        elif statement.heur_atom == HOLDS_KEY:
            if self.__is_nondet(statement, open_list):
                atom = utils.HOLDS + "("+str(heur.atom.term.arguments[0])+")"
                string = ERROR_HOLDS_H.format(self.__type, atom)
                self.__helper.raise_exception(string)
            zero = clingo.ast.Symbol(heur.location, clingo.parse_term("0"))
            heur.atom.term.arguments[1] = zero
            return DET
        # holds'
        elif statement.heur_atom == HOLDSP_KEY:
            atom = utils.HOLDSP + "("+str(heur.atom.term.arguments[0])+")"
            string = ERROR_HOLDSP_H.format(self.__type, atom)
            self.__helper.raise_exception(string)
        # open and not holds[']
        self.__add_volatile(heur.body, heur.location)
        return NONDET

    def visit_ProjectAtom(self, atom):
        string = ERROR_PROJECT.format(self.__type, str(atom))
        self.__helper.raise_exception(string)

    def visit_ProjectSignature(self, sig):
        string = ERROR_PROJECT.format(self.__type, str(sig))
        self.__helper.raise_exception(string)

    #
    # Elements
    #

    def visit_Literal(self, lit):
        atom_type = str(lit.atom.type)
        # if lit is a symbolic regular atom
        if atom_type == "SymbolicAtom" and \
           str(lit.atom.term.type) == "Function":
            # set flag
            flag = False
            if (self.in_Head and self.in_Literal_ConditionalLiteral) or \
               (self.in_Body and (
                lit.sign != NO_SIGN                         or \
                self.in_Aggregate                           or \
                self.in_TheoryAtom or self.in_BodyAggregate or \
                (self.in_ConditionalLiteral and 
                 not self.in_Literal_ConditionalLiteral))):
                flag = True
            # update graph and statements list
            term = lit.atom.term
            self.__graph.add_atom(term, self.in_Head, self.in_Body, flag)
            sig = (term.name, len(term.arguments))
            self.__statements[-1].preds.append(sig)
            if self.in_Condition:
                self.__conditions[-1].preds.append(sig)
        # if list is a special Literal, visit the children
        if atom_type not in ["Comparison", "BooleanConstant", "SymbolicAtom"]:
            self.visit_children(lit)

    # csp literals are not accepted
    def visit_CSPLiteral(self, csp):
        string = ERROR_CSPLITERAL.format(self.__type, str(csp))
        self.__helper.raise_exception(string)

    def visit_ConditionalLiteral(self, c):
        self.in_Literal_ConditionalLiteral = True
        self.visit(c.literal)
        self.in_Literal_ConditionalLiteral = False
        self.__visit_condition(c.condition, c.location)

    def visit_Aggregate(self, a):
        self.visit_children(a)

    def visit_TheoryAtom(self, th):
        for i in th.elements:
            self.__visit_condition(i.condition, th.location)

    def visit_BodyAggregate(self, b):
        for i in b.elements:
            self.__visit_condition(i.condition, b.location)

    # disjoint atoms are not accepted
    def visit_Disjoint(self, d):
        string = ERROR_DISJOINT.format(self.__type, str(d))
        self.__helper.raise_exception(string)




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

#
# IMPORTS
#

import clingo
import controller
from ..utils import printer
from ..utils import utils

# TIMING
import time
times = {}
def start_clock(clock):
    times[clock] = time.clock()
def stop_clock(clock):
    print(time.clock()-times[clock])


#
# DEFINES
#

# nodes and control
START         = "START"
START_LOOP    = "START_LOOP"
SOLVE         = "SOLVE"
SAT           = "SAT"
UNSAT         = "UNSAT"
UNKNOWN       = "UNKNOWN"
END_LOOP      = "END_LOOP"
END           = "END"
SATISFIABLE   = "SATISFIABLE"
UNSATISFIABLE = "UNSATISFIABLE"

# strings
STR_ANSWER             = "Answer: {}"
STR_OPTIMUM_FOUND      = "OPTIMUM FOUND"
STR_OPTIMUM_FOUND_STAR = "OPTIMUM FOUND *"
STR_MODEL_FOUND        = "MODEL FOUND"
STR_MODEL_FOUND_STAR   = "MODEL FOUND *"
STR_UNSATISFIABLE      = "UNSATISFIABLE"
STR_LIMIT              = "MODEL FOUND (SEARCH LIMIT)"
STR_NON_OPTIMAL        = "BETTER THAN MODEL {}"

# program names
DO_HOLDS = "do_holds"
DO_HOLDS_APPROX = "do_holds_approx"
DO_HOLDS_DELETE_BETTER = "do_holds_delete_better"
DO_HOLDS_AT_ZERO = "do_holds_at_zero"
OPEN_HOLDS = "open_holds"
VOLATILE_FACT = "volatile_fact"
VOLATILE_EXT = "volatile_external"
DELETE_MODEL = "delete_model"
DELETE_MODEL_VOLATILE = "delete_model_volatile"
DELETE_MODEL_APPROX = "delete_model_approx"
UNSAT_PRG = "unsat"
NOT_UNSAT_PRG = "not_unsat"
CMD_HEURISTIC = "cmd_heuristic"
PROJECT_APPROX = "project_approx"
PREFP = utils.PREFP
PBASE = utils.PBASE
APPROX = utils.APPROX
HEURISTIC = utils.HEURISTIC
UNSATP = utils.UNSATP
UNSATPBASE = utils.UNSATPBASE

# predicate and term names
VOLATILE      = utils.VOLATILE
MODEL         = utils.MODEL
HOLDS         = utils.HOLDS
UNSAT_ATOM    = utils.UNSAT
PREFERENCE    = utils.PREFERENCE
HOLDS_AT_ZERO = "holds_at_zero"
CSP           = "$"
MODEL_DELETE_BETTER = clingo.parse_term("delete_better")
DELETE_MODEL_VOLATILE_ATOM = "delete_model_volatile_atom"

# messages
SAME_MODEL = """\
same stable model computed twice, there is an error in the input, \
probably an incorrect preference program"""
WARNING_NO_OPTIMIZE = """WARNING: no optimize statement, \
computing non optimal stable models"""
UNKNOWN_OPTIMAL = """\nINFO: The MODELs FOUND (with SEARCH LIMIT) \
for which no BETTER MODEL was said to be found are OPTIMAL MODELS"""

#
# AUXILIARY PROGRAMS
#

TOKEN    = "##"
PROGRAMS = \
  [(DO_HOLDS_AT_ZERO,       [],"""
#show ##holds_at_zero(X) : ##""" + HOLDS + """(X,0)."""),
   (DO_HOLDS,            ["m"],"""
##""" + HOLDS + """(X,m) :- X = @get_holds()."""),
   (OPEN_HOLDS,          ["m"],"""
{ ##""" + HOLDS + """(X,m) } :- X = @get_holds_domain()."""),
   (VOLATILE_FACT, ["m1","m2"],"""
##""" + VOLATILE + """(##m(m1),##m(m2))."""),
   (VOLATILE_EXT,  ["m1","m2"],"""
#external ##""" + VOLATILE + """(##m(m1),##m(m2))."""),
   (DELETE_MODEL,           [],"""
:-     ##""" + HOLDS + """(X,0) : X = @get_holds(); 
   not ##""" + HOLDS + """(X,0) : X = @get_nholds()."""),
   (DELETE_MODEL_VOLATILE,           ["m"],"""
#external ##""" + DELETE_MODEL_VOLATILE_ATOM + """(m).
:-     ##""" + HOLDS + """(X,0) : X = @get_holds(); 
   not ##""" + HOLDS + """(X,0) : X = @get_nholds();
   not ##""" + DELETE_MODEL_VOLATILE_ATOM + """(m)."""),
   (UNSAT_PRG,     ["m1","m2"],"""
:- not ##""" + UNSAT_ATOM + """(##m(m1),##m(m2)),
       ##""" +   VOLATILE + """(##m(m1),##m(m2))."""),
   (NOT_UNSAT_PRG, ["m1","m2"],"""
:-     ##""" + UNSAT_ATOM + """(##m(m1),##m(m2)),
       ##""" +   VOLATILE + """(##m(m1),##m(m2))."""),
   (CMD_HEURISTIC,   ["v","m"],"""
#heuristic ##""" + HOLDS + """(X,0) : ##""" + PREFERENCE + 
    """(_,_,_,for(X),_). [v@0,m]"""),
   (DO_HOLDS_DELETE_BETTER,        [],"""
##""" + HOLDS + """(X,""" + str(MODEL_DELETE_BETTER) + """) :- ##""" + 
    HOLDS + """(X,0)."""),
  ]
PROGRAMS_APPROX = \
  [(DO_HOLDS_APPROX,            ["m","mm"],"""
##""" + HOLDS + """(X,m) :- X = @get_holds_approx(mm)."""),
   (DELETE_MODEL_APPROX,           ["mm"],"""
:-     ##""" + HOLDS + """(X,0) : X = @get_holds_approx(mm); 
   not ##""" + HOLDS + """(X,0) : X = @get_nholds_approx(mm)."""),
   (PROJECT_APPROX,           [],"""
#project  ##""" + HOLDS + """/2."""),
  ]
UNSAT_PREFP  = (PREFP,  ["m1","m2"], "##" + UNSAT_ATOM +"(##m(m1),##m(m2)).")
UNSAT_UNSATP = (UNSATP, ["m1","m2"], "##" + UNSAT_ATOM +"(##m(m1),##m(m2)).")

#
# Auxiliary Classes (EndException, Options)
#

class EndException(Exception):
    pass

class Options:
    pass

#
# Solver
#

class Solver:

    def __init__(self, control):
        # parameters
        self.control           = control
        self.underscores       = utils.underscores
        # strings
        self.volatile_str      = self.underscores + VOLATILE
        self.model_str         = self.underscores + MODEL
        self.holds_at_zero_str = self.underscores + HOLDS_AT_ZERO
        self.holds_str         = self.underscores + HOLDS
        self.unsat_str         = self.underscores + UNSAT_ATOM
        self.delete_str        = self.underscores + DELETE_MODEL_VOLATILE_ATOM
        # holds
        self.holds             = []
        self.nholds            = []
        # others
        self.options = Options()
        self.old_holds = None
        self.old_nholds = None
        self.shown = []
        self.solving_result = None
        self.externals  = dict()
        self.improving  = []
        self.not_improving  = set()
        self.store_nholds = True
        self.holds_domain = set()
        self.approx_opt_models = []
        self.assumptions = []
        self.last_model = None
        self.sequences = {}
        self.unknown = []
        self.unknown_non_optimal = set()
        self.grounded_delete_better = False
        # functions
        self.get_preference_parts_opt = self.get_preference_parts
        self.same_shown_function = self.same_shown_underscores
        # strings
        self.str_found      = STR_OPTIMUM_FOUND
        self.str_found_star = STR_OPTIMUM_FOUND_STAR
        # printer
        self.printer = printer.Printer()

    #
    # AUXILIARY
    #
    def get_external(self, m1, m2):
        external = self.externals.get((m1,m2))
        if external is None:
            f1 = clingo.Function(self.model_str, [m1])
            f2 = clingo.Function(self.model_str, [m2])
            external = clingo.Function(self.volatile_str, [f1, f2])
            self.externals[(m1,m2)] = external
        return external

    def get_holds(self):
        return self.holds

    def get_holds_domain(self):
        return list(self.holds_domain)

    def set_holds_domain(self):
        for i in self.control.symbolic_atoms.by_signature(self.holds_str, 2):
            if str(i.symbol.arguments[1]) == "0":
                self.holds_domain.add(i.symbol.arguments[0])
        return self.holds_domain

    def get_nholds(self):
        return self.nholds

    def get_holds_function(self, term, y):
        return clingo.Function(self.holds_str, [term, clingo.Number(y)])

    def get_unsat_function(self, term, y):
        return clingo.Function(self.unsat_str, [term, clingo.Number(y)])

    def get(self, atuple, index):
        try:
            return atuple.arguments[index.number]
        except:
            return atuple

    def length(self, atuple):
        try:
            return len(atuple.arguments)
        except:
            return 1 

    def cat(self, tuple):
        if tuple.arguments:
            return "".join([str(i) for i in tuple.arguments]).replace('"',"")
        else:
            return str(tuple)

    def get_sequence(self,name,elem):
        string = str(name)
        if string in self.sequences:
            self.sequences[string] += 1
        else:
            self.sequences[string]  = 1
        #print(self.sequences[string])
        return self.sequences[string]

    #
    # CLINGO PROXY
    #

    def add_encodings(self):
        for i in PROGRAMS:
            self.control.add(i[0], i[1], i[2].replace(TOKEN, self.underscores))
        self.control.ground([(DO_HOLDS_AT_ZERO, [])], self)

    def no_optimize(self):
        optimize = self.underscores + utils.OPTIMIZE
        for atom in self.control.symbolic_atoms.by_signature(optimize, 1):
            return False
        return True

    def print_no_optimize_warning(self):
        self.printer.print_warning(WARNING_NO_OPTIMIZE)

    def add_unsat_to_preference_program(self):
        self.control.add(UNSAT_PREFP[0], UNSAT_PREFP[1],
                         UNSAT_PREFP[2].replace(TOKEN, self.underscores))
        if self.options.preference_unsat:
            self.control.add(UNSAT_UNSATP[0], UNSAT_UNSATP[1],
                             UNSAT_UNSATP[2].replace(TOKEN, self.underscores))

    def ground_heuristic(self):
        self.control.ground([(HEURISTIC, [])], self)

    def ground_cmd_heuristic(self):
        params = [clingo.parse_term(i) for i in self.options.cmd_heuristic]
        self.control.ground([(CMD_HEURISTIC, params)], self)

    def ground_preference_base(self):
        self.control.ground([(PBASE, [])], self)

    def ground_unsatp_base(self):
        self.control.ground([(UNSATPBASE, [])], self)

    def ground_holds(self, step):
        self.control.ground([(DO_HOLDS, [step])], self)

    def get_preference_parts(self, x, y, better, volatile):
        parts = [(PREFP, [x, y])]
        if better:
            parts.append((NOT_UNSAT_PRG, [x,y]))
        else:
            parts.append((    UNSAT_PRG, [x,y]))
        if volatile:
            parts.append((VOLATILE_EXT,  [x,y]))
        else:
            parts.append((VOLATILE_FACT, [x,y]))
        return parts

    def get_preference_parts_unsatp(self, x, y, better, volatile):
        parts = self.get_preference_parts(x, y, better, volatile)
        parts.append((UNSATP, [x,y]))
        return parts[1:]

    def ground_preference_program(self, volatile):
        control, prev_step = self.control, self.step-1
        parts = self.get_preference_parts(0, prev_step, True, volatile)
        control.ground(parts, self)
        if volatile:
            if self.options.release_last:
                self.relax_previous_models()
            control.assign_external(self.get_external(0,prev_step), True)
            self.improving.append(prev_step)

    def check_errors(self):
        pr, control, u = self.printer, self.control, self.underscores
        error = False
        for atom in control.symbolic_atoms.by_signature(
            u + "_" + utils.ERROR_PRED, 1
        ):
            string = "\n" + self.cat(atom.symbol.arguments[0])
            pr.print_spec_error(string)
            error = True
        if error:
            pr.do_print("")
            raise Exception("parsing failed")

    def on_model(self, model):
        self.holds, self.nholds, self.shown = [], [], []
        for a in model.symbols(shown=True):
            if a.name == self.holds_at_zero_str:
                self.holds.append(a.arguments[0])
            else:
                self.shown.append(a)
        if self.store_nholds:
            self.nholds = list(self.holds_domain.difference(self.holds))

    def set_solving_result(self, result):
        if result.satisfiable:
            self.solving_result = SATISFIABLE
        elif result.unsatisfiable:
            self.solving_result = UNSATISFIABLE
        else:
            self.solving_result = UNKNOWN

    def solve(self):
        result = self.control.solve(assumptions=self.assumptions,
                                    on_model=self.on_model)
        self.set_solving_result(result)

    def solve_heuristic(self):
        h = []
        # gather heuristics
        for _solver in self.control.configuration.solver:
            h.append(_solver.heuristic)
            _solver.heuristic="Domain"
        # solve
        self.solve()
        # restore heuristics
        i = 0
        for _solver in self.control.configuration.solver:
            _solver.heuristic = h[i]
            i += 1

    def solve_unknown(self):
        self.solving_result = UNKNOWN

    def solve_unsat(self):
        self.solving_result = UNSATISFIABLE

    def set_control_models(self):
        solve_conf = self.control.configuration.solve
        if self.options.max_models == 0:
            solve_conf.models = 0
        else:
            solve_conf.models = self.options.max_models - self.opt_models

    def symbol2str(self,symbol):
        if symbol.name == CSP:
            return str(symbol.arguments[0]) + "=" + str(symbol.arguments[1])
        return str(symbol)

    def print_shown(self):
        self.printer.do_print(" ".join(map(self.symbol2str, self.shown)))

    def print_str_answer(self):
        self.printer.do_print(STR_ANSWER.format(self.models))

    def print_answer(self):
        self.printer.do_print(STR_ANSWER.format(self.models))
        self.printer.do_print(" ".join(map(self.symbol2str, self.shown)))

    def print_non_optimal_string(self, model):
        self.printer.do_print(STR_NON_OPTIMAL.format(model))

    def print_limit_string(self):
        self.printer.do_print(STR_LIMIT)

    def print_optimum_string(self):
        self.printer.do_print(self.str_found)

    def print_unsat(self):
        self.printer.do_print(STR_UNSATISFIABLE)

    def check_last_model(self):
        if self.old_holds  == self.holds and (
           self.old_nholds == self.nholds):
                self.printer.do_print()
                raise Exception(SAME_MODEL)
        self.old_holds  = self.holds
        self.old_nholds = self.nholds

    def same_shown(self):
        if set(self.old_shown) == set(self.shown):
            self.enumerate_flag = True
            return True
        return False

    def same_shown_underscores(self):
        u = self.underscores
        s1 = set([i for i in self.old_shown if not str(i).startswith(u)])
        s2 = set([i for i in     self.shown if not str(i).startswith(u)])
        if s1 == s2:
            self.enumerate_flag = True
            return True
        return False

    def same_shown_false(self):
        return False

    def on_model_enumerate(self, model):
        true = model.symbols(shown=True)
        self.shown = [ i for i in true if i.name != self.holds_at_zero_str ]
        # self.same_shown_function is modified by EnumerationController 
        # at controller.py
        if self.enumerate_flag or not self.same_shown_function():
            self.models     += 1
            self.opt_models += 1
            if self.options.quiet in {0,1}:
                self.print_answer()
            else:
                self.print_str_answer()
            self.printer.do_print(self.str_found_star)

    def enumerate(self):
        # models
        control    = self.control
        old_models = self.control.configuration.solve.models
        self.set_control_models()
        # assumptions
        ass  = [ (self.get_holds_function(x,0),  True) for x in self.holds ]
        ass += [ (self.get_holds_function(x,0), False) for x in self.nholds]
        # solve
        self.old_shown, self.enumerate_flag = self.shown, False
        control.solve(assumptions = ass + self.assumptions,
                         on_model = self.on_model_enumerate)
        self.shown = self.old_shown
        control.configuration.solve.models = old_models

    def relax_previous_models(self):
        for i in self.improving:
            self.control.release_external(self.get_external(0, i))
        self.improving = []

    def ground_holds_delete_better(self):
        if not self.grounded_delete_better:
            self.control.ground([(DO_HOLDS_DELETE_BETTER, [])], self)
            self.grounded_delete_better = True

    def relax_optimal_models(self):
        for x,y in self.not_improving:
            self.control.assign_external(self.get_external(x,y),False)

    def volatile_optimal_model(self, step, delete_worse, delete_better):
        if delete_worse:
            self.not_improving.add((step,0))
        if delete_better:
            self.not_improving.add((MODEL_DELETE_BETTER,step))
        for x,y in self.not_improving:        #activate
            self.control.assign_external(self.get_external(x,y),True)
        if not self.options.no_opt_improving: #reset
            self.not_improving = set()

    def handle_optimal_model(self, step, delete_model_volatile,
                             delete_worse, delete_better, volatile):
        if not delete_model_volatile:
            parts = [(DELETE_MODEL, [])]
        else:
            parts = [(DELETE_MODEL_VOLATILE, [step])]
        if delete_worse: 
            # note: get_preference_parts_opt defaults to get_preference_parts
            #       may be modified by the GeneralController
            parts += self.get_preference_parts_opt(step, 0, False, volatile)
        if delete_better:
            parts += self.get_preference_parts_opt(MODEL_DELETE_BETTER, step,
                                                   False, volatile)
        self.control.ground(parts, self)
        if volatile:
            self.volatile_optimal_model(step, delete_worse, delete_better)

    def clean_up(self):
        self.control.cleanup()

    def end(self):
        self.printer.print_stats(self.control,             self.models,
                                 self.more_models,         self.opt_models,
                                 self.options.non_optimal, self.options.stats)
        raise EndException

    #
    # ground once method
    #

    def ground_open_preference_program(self):
        parts = self.get_preference_parts(0, -1, True, True)
        parts.append((OPEN_HOLDS, [-1]))
        self.control.ground(parts, self)

    def turn_off_preference_program(self):
        self.control.assign_external(self.get_external(0,-1), False)
        self.assumptions = (
            [(self.get_holds_function(i,-1),False) for i in  self.holds_domain]
        )

    def turn_on_preference_program(self):
        self.control.assign_external(self.get_external(0,-1), True)
        self.assumptions = (
            [(self.get_holds_function(i,-1),True)  for i in  self.holds] +
            [(self.get_holds_function(i,-1),False) for i in self.nholds]
        )

    #
    # approximation
    #

    def on_model_approx(self, model):
        # for all models
        self.models += 1
        self.print_str_answer()
        # for models not proved to be optimal
        if not model.optimality_proven and model.cost!=[]:
            if self.options.quiet == 0:
                self.on_model(model)
                self.print_shown()
            elif self.options.quiet == 1 and self.options.max_models == 1:
                self.on_model(model)                            # one model
            return
        # for models proved to be optimal
        self.on_model(model)
        if self.options.quiet in {0,1}:
            self.print_shown()
        self.approx_opt_models.append(self.holds)
        self.opt_models += 1
        self.print_optimum_string()

    def get_holds_approx(self, i):
        return self.approx_opt_models[int(str(i))]

    def get_nholds_approx(self, i):
        _set = self.holds_domain.difference(self.approx_opt_models[int(str(i))])
        return list(_set)

    def solve_approx(self):
        # set the domain of holds
        self.set_holds_domain()
        # approximation programs
        self.control.ground([(APPROX, [])], self)
        for i in PROGRAMS_APPROX:
            self.control.add(i[0], i[1], i[2].replace(TOKEN, self.underscores))
        # opt_mode
        self.control.configuration.solve.opt_mode = 'optN'
        if self.options.max_models == 1:                        # one model
            self.control.configuration.solve.opt_mode = 'opt'
        # project
        if self.options.project:
            self.control.ground([(PROJECT_APPROX,[])], self)
            self.control.configuration.solve.project = 'project'
        # loop
        self.printer.do_print("Solving...")
        while True:
            # solve
            prev_opt_models, self.approx_opt_models = self.opt_models, [[]]
            self.set_control_models()
            if self.options.max_models == 1:                    # one model
                self.control.configuration.solve.models = 0
            result = self.control.solve(on_model=self.on_model_approx)
            # break if unsat or computed all or one model
            if not result.satisfiable                      or \
                self.options.max_models == self.opt_models or \
                self.options.max_models == 1:                   # one model
                break
            # add programs
            parts = []
            for mm in range(1, len(self.approx_opt_models)):
                m = mm + prev_opt_models
                parts += [(DELETE_MODEL_APPROX, [mm]),
                          (DO_HOLDS_APPROX,   [m,mm])]
                if self.options.total_order and m>1:
                    continue
                parts += self.get_preference_parts(m, 0, False, False)
                if self.options.delete_better:
                    parts += self.get_preference_parts(0, m, False, False)
            self.control.ground(parts, self)
        # end
        if self.options.max_models == 1 and result.satisfiable: # one model
            if self.options.quiet == 1:
                self.print_shown()
            self.opt_models += 1
            self.print_optimum_string()
        if self.opt_models == 0:
            self.print_unsat()
        self.more_models = True if result.satisfiable else False
        self.end()

    #
    # unknown (--improve-limit)
    #

    def computed_all(self):
        return (self.options.max_models != 0 and \
                self.opt_models == self.options.max_models)

    def print_unknown_optimal_models(self):
        self.printer.do_print(UNKNOWN_OPTIMAL)

    def enumerate_unknown(self):
        # if no unknowns, or computed all, or project: return
        if not self.unknown:
            self.more_models = False
            return
        if self.computed_all():
            return
        if self.options.project and not self.options.improve_limit[3]:
            self.print_unknown_optimal_models()
            self.opt_models += len(self.unknown)
            return
        # create boolean array representing unknown
        unknowns = [False] * (self.last_model + 1)
        for i in self.unknown:
            unknowns[i] = True
        # create holds dictionary for unknowns
        holds = {}
        for i in self.control.symbolic_atoms.by_signature(self.holds_str, 2):
            try:
                step = int(i.symbol.arguments[1].number)
                if unknowns[step]:
                    alist = holds.setdefault(step, [])
                    alist.append(i.symbol.arguments[0])
            except:
                pass
        # enumerate iterating over holds
        old = self.same_shown_function
        self.same_shown_function = self.same_shown_false
        for step in self.unknown:
            if self.computed_all():
                return
            # pre
            self.holds  = holds.get(step, [])
            self.nholds = list(self.holds_domain.difference(self.holds))
            delete_model = clingo.parse_term(
                "{}({})".format(self.delete_str, step)
            )
            self.control.assign_external(delete_model, True)
            if self.options.project:
                old = self.options.max_models
                self.options.max_models = self.opt_models + 1
            # enumerate
            self.enumerate()
            # post
            if self.options.project:
                self.options.max_models = old
            self.control.release_external(delete_model)
        self.same_shown_function = old
        self.more_models = False

    def on_model_unknown(self, model):
        self.unknown_non_optimal = set()
        atoms = model.symbols(atoms=True)
        for i in self.unknown:
            if self.get_unsat_function(MODEL_DELETE_BETTER, i) in atoms:
                self.unknown_non_optimal.add(i)

    def handle_unknown_models(self):
        # assumptions
        ass  = [ (self.get_holds_function(x,0),  True) for x in self.holds ]
        ass += [ (self.get_holds_function(x,0), False) for x in self.nholds]
        ass += [                             (x, True) for x in self.shown]
        # turn unknowns on
        for i in self.unknown:
            self.control.assign_external(
                self.get_external(MODEL_DELETE_BETTER, i), True
            )
        # solve
        if self.unknown:
            self.control.solve(assumptions = ass + self.assumptions,
                               on_model = self.on_model_unknown)
        # release non optimal, and update unknown
        update_unknown = [self.last_model]
        for i in self.unknown:
            if i in self.unknown_non_optimal:
                self.control.release_external(
                    self.get_external(MODEL_DELETE_BETTER, i)
                )
                self.control.release_external(
                    self.get_external(i, 0)
                )
                self.print_non_optimal_string(i)
            else:
                self.control.assign_external(
                    self.get_external(MODEL_DELETE_BETTER, i), False
                )
                update_unknown.append(i)
        self.unknown = update_unknown
        # update not_improving
        self.not_improving.difference_update(
            [(x,0) for x in self.unknown_non_optimal]
        )
        # add delete better for last model (without unsat constraint)
        x, y  = MODEL_DELETE_BETTER, self.last_model
        parts = [(PREFP, [x, y]), (VOLATILE_EXT,  [x,y])]
        self.control.ground(parts, self)


    #
    # OPTIONS
    #

    def set_options(self,options):
        for key,value in options.items():
            setattr(self.options,key,value)

    #
    # RUN()
    #

    def run(self):

        # controllers
        general = controller.GeneralController(self)
        optimal = controller.GeneralControllerHandleOptimal(self)
        enumeration = controller.EnumerationController(self)
        # MethodController
        if self.options.solving_mode == "approx":
            method = controller.ApproxMethodController(self)
        elif self.options.solving_mode == "heuristic":
            method = controller.HeurMethodController(self)
        else:
            if self.options.ground_once:
                method = controller.GroundOnceMethodController(self)
            else:
                method = controller.GroundManyMethodController(self)
        if self.options.improve_limit is not None:
            method = controller.ImproveLimitController(self, method)

        # loop
        try:
            # START
            general.start()
            optimal.start()
            method.start()
            self.printer.do_print("Solving...")
            while True:
                # START_LOOP
                method.start_loop()
                # SOLVE
                method.solve()
                if self.solving_result == SATISFIABLE:
                    # SAT
                    general.sat()
                    optimal.sat()
                elif self.solving_result == UNSATISFIABLE:
                    # UNSAT
                    general.unsat()
                    method.unsat()
                    enumeration.unsat()
                    optimal.unsat()
                elif self.solving_result == UNKNOWN:
                    # UNKNOWN
                    general.unknown()
                    method.unsat()
                    optimal.unknown()
                # END_LOOP
                general.end_loop()
        except RuntimeError as e:
            self.printer.print_error("ERROR (clingo): {}".format(e))
        except EndException as e:
            # END
            pass





# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

import clingo
import clingo.ast
from collections import namedtuple
from ..utils import utils
from ..utils import printer


#
# DEFINES
#

# predicate names
VOLATILE   = utils.VOLATILE
HOLDS      = utils.HOLDS
HOLDSP     = utils.HOLDSP
PREFERENCE = utils.PREFERENCE
OPTIMIZE   = utils.OPTIMIZE
UNSAT      = utils.UNSAT

# others
MODEL = utils.MODEL
EMPTY = ""
M1    = "m1"
M2    = "m2"
M1_M2 = "m1_m2"
SHOW  = "show"
EDGE  = "edge"
ZERO  = "zero"

NO_SIGN = utils.NO_SIGN

#
# CLASSES
#


PredicateInfo = namedtuple('PredicateInfo','name underscores ems arity')


class Helper:

    underscores, m1, m2            = None, None, None
    simple_m1, simple_m2, volatile = None, None, None
    unsat, show, edge              = None, None, None
    zero                           = None

    def __init__(self):
        if self.underscores is not None:
            return
        Helper.underscores = utils.underscores
        term = "{}({})".format(self.underscore(MODEL), self.underscore(M1))
        Helper.m1 = clingo.parse_term(term)
        term = "{}({})".format(self.underscore(MODEL), self.underscore(M2))
        Helper.m2 = clingo.parse_term(term)
        term = "{}".format(self.underscore(M1)) # for holds
        Helper.simple_m1 = clingo.parse_term(term)
        term = "{}".format(self.underscore(M2)) # for holds
        Helper.simple_m2 = clingo.parse_term(term)
        Helper.zero = clingo.parse_term("0")
        Helper.volatile = self.underscore(VOLATILE)
        Helper.unsat = self.underscore(UNSAT)
        Helper.show = self.underscore(SHOW)
        Helper.edge = self.underscore(EDGE)

    #TODO: collect many messages, and output them together
    def raise_exception(self, string):
        printer.Printer().print_error(string)
        raise Exception("parsing failed")

    def underscore(self,x):
        return self.underscores + x

    def get_ems(self, loc, ems):
        if   ems == M1:
            return [clingo.ast.Symbol(loc, self.simple_m1)]
        elif ems == M2:
            return [clingo.ast.Symbol(loc, self.simple_m2)]
        elif ems == M1_M2:
            return [clingo.ast.Symbol(loc, self.m1),
                    clingo.ast.Symbol(loc, self.m2)]
        elif ems == ZERO:
            return [clingo.ast.Symbol(loc, self.zero)]
        else:
            return []

    def get_volatile_atom(self, loc):
        ems = self.get_ems(loc, M1_M2)
        fun = clingo.ast.Function(loc, self.volatile, ems, False)
        return clingo.ast.Literal(loc, NO_SIGN, clingo.ast.SymbolicAtom(fun))


class Visitor:

    def __init__(self):
        for i in dir(self):
            if i.startswith("visit_"):
                setattr(self, "in_"+i[6:], False)

    def finish(self):
        pass

    def visit_children(self, x, *args, **kwargs):
        for key in x.child_keys:
            self.visit(getattr(x, key), *args, **kwargs)
        return x

    def visit(self, x, *args, **kwargs):
        if isinstance(x, clingo.ast.AST):
            attr = "visit_" + str(x.type)
            if hasattr(self, attr):
                setattr(x, "in_"+attr, True)  # added
                getattr(self, attr)(x, *args, **kwargs)
                setattr(x, "in_"+attr, False) # added
            else:
                self.visit_children(x, *args, **kwargs)
        elif isinstance(x, list):
            for y in x:
                self.visit(y, *args, **kwargs)
        elif x is None:
            pass
        else:
            raise TypeError("unexpected type")


class TermTransformer: # ABSTRACT CLASS

    def __init__(self):
        self.predicates_info = None
        self.default = None
        self.helper = Helper()

    # TO BE DEFINED BY SUBCLASSES
    def set_predicates_info(self):
        pass

    def __get_predicate_info(self, name, arity):
        predicate_info = self.predicates_info.get((name, arity))
        if predicate_info is not None:
            return predicate_info
        return self.default

    # pre: set_predicates_info() was called before
    def transform_function(self, term):
        # get info, change name, add underscores, and update arguments
        predicate_info = self.__get_predicate_info(term.name,
                                                   len(term.arguments))
        if predicate_info.name:
            term.name = predicate_info.name
        term.name = self.helper.underscore(term.name)
        term.name = "_"*predicate_info.underscores + term.name
        term.arguments += self.helper.get_ems(term.location, predicate_info.ems)

    # pre: set_predicates_info() was called before
    def transform_signature(self, sig):
        # get info, change name, add underscores, and update arity
        predicate_info = self.__get_predicate_info(sig.name, sig.arity)
        if predicate_info.name:
            sig.name = predicate_info.name
        sig.name = self.helper.underscore(sig.name)
        sig.name = "_"*predicate_info.underscores + sig.name
        sig.arity += predicate_info.arity

    # returns a Function reifying term with name
    def reify_term(self, term, name):
        return clingo.ast.Function(term.location, name, [term], False)

    # extends a Function with ems
    def extend_function(self, func, ems):
        func.arguments += self.helper.get_ems(func.location, ems)




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

from __future__ import print_function
import clingo
import sys
from ..utils import utils
from ..utils import printer
import preference
import basic

#
# DEFINES
#

EMPTY      = utils.EMPTY

# programs
BASE         = utils.BASE
SPEC         = utils.SPEC
GENERATE     = utils.GENERATE
PREFP        = utils.PREFP
APPROX       = utils.APPROX
HEURISTIC    = utils.HEURISTIC
UNSATP       = utils.UNSATP
CONSTANTS_NB = utils.CONSTANTS_NB

# underscores
U_PREFP     = utils.U_PREFP
U_APPROX    = utils.U_APPROX
U_HEURISTIC = utils.U_HEURISTIC
U_UNSATP    = utils.U_UNSATP

# predicate names
DOM        = utils.DOM
GEN_DOM    = utils.GEN_DOM
PREFERENCE = utils.PREFERENCE
SHOW       = utils.SHOW
OPTIMIZE   = utils.OPTIMIZE
ERROR_PRED = utils.ERROR_PRED
WARN_PRED  = utils.WARN_PRED

# error checking
CHECK_SPEC = """
% P names X
##names(P,X) :- ##preference(P,_,_,name(X),_).

% transitive closure of names
##tr_names(X,Y) :- ##names(X,Y).
##tr_names(X,Y) :- ##names(X,Z), ##tr_names(Z,Y).

%
% errors
%

% two types for one preference name
##error((A,B,C,D)):- ##preference(P,T1), ##preference(P,T2), T1 != T2, 
  A = "preference:", 
  B = P, 
  C = ": error: preference specification error, ", 
  D = "preference name has more than one type".

% naming non existent statement
##error((A,B,C,D,E)):- ##names(X,Y), not ##preference(Y,_), 
  A = "preference:", 
  B = X, 
  C = ": error: preference specification error, ", 
  D = "naming non existent preference statement ",
  E = Y.

% naming loops
##error((A,B,C,D)):- ##tr_names(X,X),
  A = "preference:", 
  B = X, 
  C = ": error: preference specification error, ", 
  D = "naming loop".

% optimizing non existent statement
##error((A,B,C,D)):- ##optimize(X), not ##preference(X,_),
  A = "optimize:", 
  B = X, 
  C = ": error: preference specification error, ", 
  D = "optimizing non existent preference statement".

% many optimize statements
##error((A,B,C,D)):- ##optimize(X), 2 { ##optimize(Y) }, 
  A = "optimize:", 
  B = X, 
  C = ": error: preference specification error, ", 
  D = "many optimize statements".

%
% warnings
%

% no optimize statements
##warning((A,B)):- { ##optimize(Y) } 0, 
  A = "WARNING: no optimize statement, ",
  B = "computing non optimal stable models".

"""
AVOID_WARNINGS = """#program """ + utils.WARNINGS + """.

% avoid clingo warnings
##preference(A,B,C, for(D),E) :- #false, ##preference(A,B,C, for(D),E).
##preference(A,B,C,name(D),E) :- #false, ##preference(A,B,C,name(D),E).
##optimize(X) :- #false, ##optimize(X).
##preference(X,Y) :- #false, ##preference(X,Y).
##unsat(X,Y) :- #false, ##unsat(X,Y).
##false :- #false, ##false.
"""

# error printing
ERROR_SPEC = "error: preference specification error, "
ERROR_NON_DOMAIN = ERROR_SPEC + "the body contains non domain atoms\n"
ERROR_PREF_NON_DOMAIN  = "preference:{}: "            + ERROR_NON_DOMAIN
ERROR_ELEM_NON_DOMAIN  = "preference:{}:element:{}: " + ERROR_NON_DOMAIN
ERROR_OPT_NON_DOMAIN   = "optimize:{}: "             + ERROR_NON_DOMAIN
ERROR_NO_PREF_PROGRAM  = "preference:{}: " + ERROR_SPEC + """\
preference type '{}' has no preference program\n"""
ERROR_NO_HEURISTIC_PROGRAM  = "preference:{}: " + ERROR_SPEC + """\
preference type '{}' has no heuristic program\n"""
ERROR_NO_APPROX_PROGRAM  = "preference:{}: " + ERROR_SPEC + """\
preference type '{}' has no approximation program\n"""
ERROR_NO_UNSATP_PROGRAM  = "preference:{}: " + ERROR_SPEC + """\
preference type '{}' has no """ + UNSATP + """ program\n"""


class BuilderProxy:

    def __init__(self, builder):
        self.builder = builder
        self.printer = printer.Printer()

    def add(self, statement):
        self.printer.do_print(statement)
        self.builder.add(statement)


class Parser:

    def __init__(self, control, programs, options):
        self.__control = control
        self.__programs = programs
        self.__options = options
        self.__underscores = utils.underscores

    def __add_and_ground(self, name, params, string, list):
        capturer = utils.Capturer(sys.stderr)
        try:
            self.__control.add(name,params,string)
            self.__control.ground(list)
        finally:
            s = capturer.translate_error(self.__programs, name, capturer.read())
            capturer.close()
            if s != "":
                printer.Printer().print_error_string(s)

    def print_basic_programs(self, types):
        for name in [BASE, GENERATE, SPEC]:
            for type, program in self.__programs[name].items():
                if type in types:
                    params = "({})".format(type) if type != "" else ""
                    p = printer.Printer()
                    p.do_print("#program {}{}.".format(name,params))
                    p.do_print(program.get_string())

    def do_base(self):
        options, control = self.__options, self.__control
        programs = self.__programs
        # constants
        constants = options['constants'].items()
        old = [ key                      for key, value in constants ]
        new = [ clingo.parse_term(value) for key, value in constants ]
        # add and ground
        string =  programs[BASE][""].get_string()
        self.__add_and_ground(BASE,old,string,[(BASE,new)])
        string =  programs[GENERATE][""].get_string()
        self.__add_and_ground(GENERATE,[],string,[(GENERATE,[])])

    def get_domains(self):
        out, control, u = "", self.__control, self.__underscores
        for atom in control.symbolic_atoms.by_signature(u+GEN_DOM,2):
            args = atom.symbol.arguments
            for atom2 in control.symbolic_atoms.by_signature(str(args[0]),
                                                             int(str(args[1]))):
                out += u + DOM + "(" + str(atom2.symbol) + ").\n"
        return out

    def __cat(self, tuple):
        if tuple.arguments:
            return "".join([str(i) for i in tuple.arguments]).replace('"',"")
        else:
            return str(tuple)

    def __non_domain_message(self, atom, predicate):
        # preference statement
        if predicate == (PREFERENCE,2):
            u = self.__underscores
            arg1 = str(atom.symbol.arguments[0]).replace(u, "", 1)
            return ERROR_PREF_NON_DOMAIN.format(arg1)
        # preference element
        elif predicate == (PREFERENCE,5):
            arg1 = str(atom.symbol.arguments[0])
            try: # careful here
                arg2 = str(atom.symbol.arguments[1].arguments[0].arguments[1])
            except Exception as e:
                arg2 = str(atom.symbol.arguments[1])
            return ERROR_ELEM_NON_DOMAIN.format(arg1, arg2)
        # optimize statement
        else:
            arg1 = str(atom.symbol.arguments[0])
            return ERROR_OPT_NON_DOMAIN.format(arg1)

    def __program_exists(self, atom):
        # get programs
        programs = [(PREFERENCE, ERROR_NO_PREF_PROGRAM)]
        if self.__options['solving_mode'] == 'heuristic':
            programs.append((HEURISTIC, ERROR_NO_HEURISTIC_PROGRAM))
        elif self.__options['solving_mode'] == 'approx':
            programs.append((APPROX, ERROR_NO_APPROX_PROGRAM))
        if self.__options['preference_unsat']:
            programs.append((UNSATP, ERROR_NO_UNSATP_PROGRAM))
        # arguments of atom
        arg1 = str(atom.symbol.arguments[0])
        arg2 = str(atom.symbol.arguments[1])
        # loop
        ok = True
        for program, error in programs:
            if arg2 not in self.__programs[program]:
                printer.Printer().print_spec_error(error.format(arg1, arg2))
                ok = False
        return ok

    def do_spec(self):

        control, programs = self.__control, self.__programs
        u, options = self.__underscores, self.__options

        # specification
        constants = options['constants'].items()
        old = [ key                      for key, value in constants ]
        new = [ clingo.parse_term(value) for key, value in constants ]
        string  = programs[SPEC][""].get_string() 
        if options['check']:
            string += CHECK_SPEC.replace("##",u)
        string += AVOID_WARNINGS.replace("##",u)
        self.__add_and_ground(SPEC,old,string,[(SPEC,new)])

        pr = printer.Printer()
        errors = False

        if options['check']:
            # get specification warnings and errors
            for atom in control.symbolic_atoms.by_signature(u+WARN_PRED, 1):
                string = self.__cat(atom.symbol.arguments[0]) + "\n"
                pr.print_spec_warning(string)
            for atom in control.symbolic_atoms.by_signature(u+ERROR_PRED, 1):
                string = self.__cat(atom.symbol.arguments[0]) + "\n"
                pr.print_spec_error(string)
                errors = True
            # get non domain errors
            for i in [(PREFERENCE,2),(PREFERENCE,5),(OPTIMIZE,1)]:
                ui0 = u + i[0]
                for atom in control.symbolic_atoms.by_signature(ui0, i[1]):
                    if not atom.is_fact:
                        pr.print_spec_error(self.__non_domain_message(atom, i))
                        errors = True

        # get preference types, and test for corresponding programs
        out = set()
        upreference = u + PREFERENCE
        for atom in control.symbolic_atoms.by_signature(upreference,2):
            out.add(str(atom.symbol.arguments[1]))
            if options['check']:
                ok = self.__program_exists(atom)
                if not ok:
                    errors = True

        # if errors
        if errors:
            raise Exception("parsing failed")

        return out

    def add_show(self,types):

        # decide if adding #shows to the base, and set 'show_underscores'
        self.__options['show_underscores'], add_base_show = False, True
        for i in self.__options['show']:
            if (i[0]==BASE and i[1]==EMPTY):
                add_base_show = False
            elif (i[0]==PREFERENCE and i[1] in types):
                self.__options['show_underscores'] = True

        if not add_base_show: 
            return

        # else add #show for atoms in base
        # NOTE: without #show statements, csp variables are not shown
        show = "\n".join(["#show " + ("" if x[2] else "-") + x[0] + "/" +
                          str(x[1]) + "."
                          for x in self.__control.symbolic_atoms.signatures if
                          not str(x[0]).startswith(self.__underscores)])
        show += "#show."
        self.__add_and_ground(SHOW,[],show,[(SHOW,[])])

    def add_constants_nonbase(self):
        if not self.__options['constants_nb']:
            return
        constants_nb = self.__options['constants_nb'].items()
        program = " ".join("#const {}={}.".format(x,y) for x,y in constants_nb)
        self.__control.add(CONSTANTS_NB,[],program)
        self.__control.ground([(CONSTANTS_NB,[])])

    def add_programs(self, types, builder):
        # visitors
        constants = self.__options['constants_nb']
        v = preference.PreferenceProgramVisitor(builder, PREFP,
                                                U_PREFP, constants)
        visitors = [(PREFP, v)]
        if self.__options['solving_mode'] == 'approx':
            v = basic.BasicProgramVisitor(builder, APPROX, U_APPROX, constants)
            visitors.append((APPROX,v))
        elif self.__options['solving_mode'] == 'heuristic':
            v = basic.HeuristicProgramVisitor(builder, HEURISTIC,
                                              U_HEURISTIC, constants)
            visitors.append((HEURISTIC,v))
        if self.__options['preference_unsat']:
            v = preference.PreferenceProgramVisitor(builder, UNSATP,
                                                    U_UNSATP, constants)
            visitors.append((UNSATP,v))
        # add programs
        for name, visitor in visitors:
            for type_, program in self.__programs[name].items():
                if type_ in types:
                    s = "#program " + name + ".\n" + program.get_string()
                    clingo.parse_program(s, lambda x: visitor.visit(x))
            visitor.finish()

    def parse(self):

        # ground base program
        self.do_base()

        # get domains for the specification
        self.__programs[SPEC][""].extend_string(self.get_domains())

        # ground specification and get preference types
        types = self.do_spec()
        types.add("") # add basic case

        # add #show statements if needed (CSP variables are not shown)
        self.add_show(types)

        # add nonbase constants
        self.add_constants_nonbase()

        # option --print-programs
        if self.__options['print-programs']:
            self.print_basic_programs(types)
            with self.__control.builder() as _builder:
                builder = BuilderProxy(_builder)
                self.add_programs(types, builder)
            raise utils.SilentException() # end

        # translate and add the rest of the programs
        with self.__control.builder() as builder:
            self.add_programs(types, builder)




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys

PYTHON = sys.executable
ASPRIN = PYTHON + " " + str(os.path.join(os.path.dirname(os.path.realpath(
                            __file__)),"..","..","asprin.py")) + " "

class Test:

    def __init__(self, string, options):
        self.command = None
        self.satisfiable = False
        self.unsatisfiable = False
        self.error = False
        self.answers = None
        self.parse(string, options)
        self.string = string

    def parse(self, string, options):
        line, last = 1, ""
        self.answers = [] 
        for i in string.splitlines():
            if line == 1:
                self.command = i[1:]
                self.command = self.command.replace('asprin',
                                                    ASPRIN+' '.join(options),1)
            elif line == 2:
                match = re.match(r'%( )*SATISFIABLE', i)
                self.satisfiable = True if match else False
                match = re.match(r'%( )*UNSATISFIABLE', i)
                self.unsatisfiable = True if match else False
                match = re.match(r'%( )*ERROR', i)
                self.error = True if match else False
            else:
                match = re.match(r'%( )*((OPTIMUM)|(MODEL)) FOUND.*', i)
                if match:
                    answer = last[1:].split(' ')
                    answer.sort()
                    answer = " ".join(answer)
                    self.answers.append(answer)
            line, last = line+1, i
        self.answers.sort()
        if not self.satisfiable and not self.unsatisfiable and not self.error:
            print("PARSING ERROR in class Test")

    def __repr__(self):
        return self.command + "\n" + self.string
        if self.command:
            out = "\ncommand = \"{}\"\n".format(self.command)
        else:
            out = ""
        if self.satisfiable:
            out += "SATISFIABLE\n"
        if self.unsatisfiable:
            out += "UNSATISFIABLE\n"
        if self.error:
            out += "ERROR\n"
        answers = "\n".join(self.answers)
        out += "Answers:\n{}".format(answers)
        return out


class Result(Test):

    def __init__(self, string):
        self.count = 0
        Test.__init__(self, string, [])

    def __repr__(self):
        return self.string

    def parse(self, string, options):
        line, last = 1, ""
        self.answers = []
        for i in string.splitlines():
            match = re.match(r'.*ERROR.*', i)
            if match:
                self.error = True
            match = re.match(r'.*UNSATISFIABLE.*', i)
            if match:
                self.unsatisfiable = True
            match = re.match(r'.*((OPTIMUM)|(MODEL)) FOUND.*', i)
            match_limit = re.match(r'.*MODEL FOUND \(.*', i)
            if match and not match_limit:
                self.satisfiable = True
                answer = last.split(' ')
                answer.sort()
                answer = " ".join(answer)
                self.answers.append(answer)
            line, last = line+1, i
        self.answers.sort()
        if self.satisfiable:   self.count += 1
        if self.unsatisfiable: self.count += 1
        if self.error:         self.count += 1

    def __print_error(self, test, message):
        print("#############################################################")
        print("*************************************************************")
        print("ERROR: " + message)
        print("*************************************************************")
        print("EXPECTED:")
        print("*************************************************************")
        print(test)
        print("*************************************************************")
        print("RESULT:")
        print("*************************************************************")
        print(self)
        print("#############################################################\n")

    def compare(self, test):
        if self.count != 1:
            msg = "ERROR, UNSATISFIABLE or OPTIMUM FOUND messages"
            self.__print_error(test, "parsing error, 0 or many " + msg)
        elif self.satisfiable != test.satisfiable:
            self.__print_error(test, "result is satisfiable")
        elif self.unsatisfiable != test.unsatisfiable:
            self.__print_error(test, "result is unsatisfiable")
        elif self.error != test.error:
            self.__print_error(test, "result gives an error")
        elif "\n".join(self.answers) != "\n".join(test.answers):
            self.__print_error(test, "different answers")
        else:
            return False
        return True

